# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

"""Generic util functions used in the code"""

import base64
from filelock import FileLock
from dataclasses import dataclass
from datetime import date, datetime, time, timezone
from decimal import Decimal
from enum import Enum
from heapq import heappush
import ipaddress
import json
import logging
import mimetypes
import os
import pathlib
from pathlib import Path
import re
import socket
from typing import Any, IO, Union, List, Optional
from urllib.parse import urlparse
from urllib.request import urlopen
import uuid

import dateutil.parser
from babel.support import Translations
from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2.exceptions import TemplateNotFound
from requests import Session
from requests.structures import CaseInsensitiveDict
from shapely.geometry import (
    box,
    Polygon,
    mapping as geom_to_geojson,
)
import yaml

from pygeoapi import __version__
from pygeoapi import l10n
from pygeoapi.models import config as config_models
from pygeoapi.plugin import load_plugin, PLUGINS
from pygeoapi.provider import get_provider_default

LOGGER = logging.getLogger(__name__)

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

THISDIR = Path(__file__).parent.resolve()
RESOURCESDIR = THISDIR / 'resources'
TEMPLATESDIR = THISDIR / 'templates'
DEFINITIONSDIR = RESOURCESDIR / 'definitions'
SCHEMASDIR = RESOURCESDIR / 'schemas'


mimetypes.add_type('text/plain', '.yaml')
mimetypes.add_type('text/plain', '.yml')


def dategetter(date_property: str, collection: dict) -> str:
    """
    Attempts to obtain a date value from a collection.

    :param date_property: property representing the date
    :param collection: dictionary to check within

    :returns: `str` (ISO8601) representing the date (allowing
               for an open interval using null)
    """

    value = collection.get(date_property)

    if value is None or isinstance(value, str):
        return value
    else:
        return value.isoformat()


def get_typed_value(value: str) -> Union[bool, float, int, str]:
    """
    Derive true type from data value

    :param value: value

    :returns: value as a native Python data type
    """

    try:
        if '.' in value:  # float?
            value2 = float(value)
        elif len(value) > 1 and value.startswith('0'):
            value2 = value
        elif value.lower() in ['true', 'false']:
            value2 = str2bool(value)
        else:  # int?
            value2 = int(value)
    except ValueError:  # string (default)?
        value2 = value

    return value2


def yaml_load(fh: IO) -> dict:
    """
    serializes a YAML files into a pyyaml object

    :param fh: file handle

    :returns: `dict` representation of YAML
    """

    # # support environment variables in config
    # # https://stackoverflow.com/a/55301129

    env_matcher = re.compile(
        r'.*?\$\{(?P<varname>\w+)(:-(?P<default>[^}]*))?\}')

    def env_constructor(loader, node):
        result = ""
        current_index = 0
        raw_value = node.value
        for match_obj in env_matcher.finditer(raw_value):
            groups = match_obj.groupdict()
            varname_start = match_obj.span('varname')[0]
            result += raw_value[current_index:(varname_start-2)]
            if (var_value := os.getenv(groups['varname'])) is not None:
                result += var_value
            elif (default_value := groups.get('default')) is not None:
                result += default_value
            else:
                raise EnvironmentError(
                    f'Could not find the {groups["varname"]!r} environment '
                    f'variable'
                )
            current_index = match_obj.end()
        else:
            result += raw_value[current_index:]
        return get_typed_value(result)

    class EnvVarLoader(yaml.SafeLoader):
        pass

    EnvVarLoader.add_implicit_resolver('!env', env_matcher, None)
    EnvVarLoader.add_constructor('!env', env_constructor)
    return yaml.load(fh, Loader=EnvVarLoader)


def get_api_rules(config: dict) -> config_models.APIRules:
    """ Extracts the default API design rules from the given configuration.

    :param config:  Current pygeoapi configuration (dictionary).
    :returns:       An APIRules instance.
    """
    rules = config['server'].get('api_rules') or {}
    rules.setdefault('api_version', __version__)
    return config_models.APIRules.create(**rules)


def get_base_url(config: dict) -> str:
    """ Returns the full pygeoapi base URL. """
    rules = get_api_rules(config)
    return url_join(config['server']['url'], rules.get_url_prefix())


def yaml_dump(dict_: dict, destfile: str) -> bool:
    """
    Dump dict to YAML file

    :param dict_: `dict` to dump
    :param destfile: destination filepath

    :returns: `bool`
    """

    def path_representer(dumper, data):
        return dumper.represent_scalar(u'tag:yaml.org,2002:str', str(data))

    yaml.add_multi_representer(pathlib.PurePath, path_representer)

    def datetime_representer(dumper, data: datetime):
        if data.tzinfo is None:
            data = data.replace(tzinfo=timezone.utc)
        else:
            data = data.astimezone(timezone.utc)
        value = data.strftime("%Y-%m-%dT%H:%M:%SZ")

        # timestamp in a specified format, without string quotes
        return dumper.represent_scalar(u'tag:yaml.org,2002:timestamp', value)

    yaml.add_representer(datetime, datetime_representer)

    lock = FileLock(f'{destfile}.lock')

    with lock:
        LOGGER.debug('Dumping YAML document')
        with open(destfile, 'wb') as fh:
            yaml.dump(dict_, fh, sort_keys=False, encoding='utf8',
                      indent=4, default_flow_style=False,
                      allow_unicode=True)

    return True


def str2bool(value: Union[bool, str]) -> bool:
    """
    helper function to return Python boolean
    type (source: https://stackoverflow.com/a/715468)

    :param value: value to be evaluated

    :returns: `bool` of whether the value is boolean-ish
    """

    value2 = False

    if isinstance(value, bool):
        value2 = value
    else:
        value2 = value.lower() in ('yes', 'true', 't', '1', 'on')

    return value2


def to_json(dict_: dict, pretty: bool = False) -> str:
    """
    Serialize dict to JSON

    :param dict_: `dict` of JSON representation
    :param pretty: `bool` of whether to prettify JSON (default is `False`)

    :returns: JSON string representation
    """

    if pretty:
        indent = 4
    else:
        indent = None

    LOGGER.debug('Dumping JSON')
    json_dump = json.dumps(dict_, default=json_serial, indent=indent,
                           separators=(',', ':'))

    LOGGER.debug('Removing < and >')
    json_dump = json_dump.replace('<', '&lt').replace('>', '&gt')

    return json_dump


def format_datetime(value: str, format_: str = DATETIME_FORMAT) -> str:
    """
    Parse datetime as ISO 8601 string; re-present it in particular format
    for display in HTML

    :param value: `str` of ISO datetime
    :param format_: `str` of datetime format for strftime

    :returns: string
    """

    if not isinstance(value, str) or not value.strip():
        return ''

    return dateutil.parser.isoparse(value).strftime(format_)


def get_current_datetime(tz: timezone = timezone.utc,
                         format_: str = DATETIME_FORMAT) -> str:
    return datetime.now(tz).strftime(format_)


def file_modified_iso8601(filepath: Path) -> str:
    """
    Provide a file's ctime in ISO8601

    :param filepath: path to file

    :returns: string of ISO8601
    """

    return datetime.fromtimestamp(
        os.path.getctime(filepath)).strftime('%Y-%m-%dT%H:%M:%SZ')


def human_size(nbytes: int) -> str:
    """
    Provides human readable file size

    source: https://stackoverflow.com/a/14996816

    :param nbytes: int of file size (bytes)
    :param units: list of unit abbreviations

    :returns: string of human readable filesize
    """

    suffixes = ['B', 'K', 'M', 'G', 'T', 'P']

    i = 0

    while nbytes >= 1024 and i < len(suffixes)-1:
        nbytes /= 1024.
        i += 1

    if suffixes[i] == 'K':
        f = str(int(nbytes)).rstrip('0').rstrip('.')
    elif suffixes[i] == 'B':
        return nbytes
    else:
        f = f'{nbytes:.1f}'.rstrip('0').rstrip('.')

    return f'{f}{suffixes[i]}'


def format_duration(start: str, end: str = None) -> str:
    """
    Parse a start and (optional) end datetime as ISO 8601 strings, calculate
    the difference, and return that duration as a string.

    :param start: `str` of ISO datetime
    :param end: `str` of ISO datetime, defaults to `start` for a 0 duration

    :returns: string
    """

    if not isinstance(start, str) or not start.strip():
        return ''
    end = end or start
    duration = dateutil.parser.isoparse(end) - dateutil.parser.isoparse(start)
    return str(duration)


def get_path_basename(urlpath: str) -> str:
    """
    Helper function to derive file basename

    :param urlpath: URL path

    :returns: string of basename of URL path
    """

    return Path(urlpath).name


def json_serial(obj: Any) -> str:
    """
    helper function to convert to JSON non-default
    types (source: https://stackoverflow.com/a/22238613)

    :param obj: `object` to be evaluated

    :returns: JSON non-default type to `str`
    """

    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        try:
            LOGGER.debug('Returning as UTF-8 decoded bytes')
            return obj.decode('utf-8')
        except UnicodeDecodeError:
            LOGGER.debug('Returning as base64 encoded JSON object')
            return base64.b64encode(obj)
    elif isinstance(obj, Decimal):
        return float(obj)
    elif type(obj).__name__ in ['int32', 'int64']:
        return int(obj)
    elif type(obj).__name__ in ['float32', 'float64']:
        return float(obj)
    elif isinstance(obj, l10n.Locale):
        return l10n.locale2str(obj)
    elif isinstance(obj, (pathlib.PurePath, Path)):
        return str(obj)
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    else:
        msg = f'{obj} type {type(obj)} not serializable'
        LOGGER.error(msg)
        raise TypeError(msg)


def is_url(urlstring: str) -> bool:
    """
    Validation function that determines whether a candidate URL should be
    considered a URI. No remote resource is obtained; this does not check
    the existence of any remote resource.

    :param urlstring: `str` to be evaluated as candidate URL.

    :returns: `bool` of whether the URL looks like a URL.
    """
    try:
        result = urlparse(urlstring)
        return bool(result.scheme and result.netloc)
    except ValueError:
        return False


def render_j2_template(config: dict, tpl_config: dict, template: Path,
                       data: dict, locale_: str = None) -> str:
    """
    render Jinja2 template

    :param config: dict of configuration
    :param tpl_config: dict of template configuration
    :param template: template (relative path)
    :param data: dict of data
    :param locale_: the requested output Locale

    :returns: string of rendered template
    """

    template_paths = [TEMPLATESDIR, '.']

    locale_dir = config['server'].get('locale_dir', 'locale')
    LOGGER.debug(f'Locale directory: {locale_dir}')

    try:
        templates = tpl_config['path']
        template_paths.insert(0, templates)
        LOGGER.debug(f'using custom templates: {templates}')
    except (KeyError, TypeError):
        LOGGER.debug(f'using default templates: {TEMPLATESDIR}')

    env = Environment(loader=FileSystemLoader(template_paths),
                      extensions=['jinja2.ext.i18n'],
                      autoescape=select_autoescape())

    env.filters['to_json'] = to_json
    env.filters['format_datetime'] = format_datetime
    env.filters['format_duration'] = format_duration
    env.filters['human_size'] = human_size
    env.globals.update(to_json=to_json)

    env.filters['get_path_basename'] = get_path_basename
    env.globals.update(get_path_basename=get_path_basename)

    env.filters['get_breadcrumbs'] = get_breadcrumbs
    env.globals.update(get_breadcrumbs=get_breadcrumbs)

    env.filters['filter_dict_by_key_value'] = filter_dict_by_key_value
    env.globals.update(filter_dict_by_key_value=filter_dict_by_key_value)

    translations = Translations.load(locale_dir, [locale_])
    env.install_gettext_translations(translations)

    try:
        template = env.get_template(template)
    except TemplateNotFound:
        LOGGER.debug(f'template {template} not found')
        template_paths.remove(templates)
        template = env.get_template(template)

    return template.render(config=l10n.translate_struct(config, locale_, True),
                           data=data, locale=locale_, version=__version__)


def get_mimetype(filename: str) -> str:
    """
    helper function to return MIME type of a given file

    :param filename: filename (with extension)

    :returns: MIME type of given filename
    """

    return mimetypes.guess_type(filename)[0]


def get_breadcrumbs(urlpath: str) -> list:
    """
    helper function to make breadcrumbs from a URL path

    :param urlpath: URL path

    :returns: `list` of `dict` objects of labels and links
    """

    links = []

    tokens = urlpath.split('/')

    s = ''
    for t in tokens:
        if s:
            s += '/' + t
        else:
            s = t
        links.append({
            'href': s,
            'title': t,
        })

    return links


def filter_dict_by_key_value(dict_: dict, key: str, value: str) -> dict:
    """
    helper function to filter a dict by a dict key

    :param dict_: ``dict``
    :param key: dict key
    :param value: dict key value

    :returns: filtered ``dict``
    """

    return {k: v for (k, v) in dict_.items() if v[key] == value}


class ProcessExecutionMode(Enum):
    sync_execute = 'sync-execute'
    async_execute = 'async-execute'


class RequestedProcessExecutionMode(Enum):
    wait = 'wait'
    respond_async = 'respond-async'


class RequestedResponse(Enum):
    raw = 'raw'
    document = 'document'


class JobStatus(Enum):
    """
    Enum for the job status options specified in the WPS 2.0 specification
    """

    #  From the specification
    accepted = 'accepted'
    running = 'running'
    successful = 'successful'
    failed = 'failed'
    dismissed = 'dismissed'


@dataclass(frozen=True)
class Subscriber:
    """
    Store subscriber URLs as defined in:

    https://schemas.opengis.net/ogcapi/processes/part1/1.0/openapi/schemas/subscriber.yaml  # noqa
    """

    success_uri: str
    in_progress_uri: Optional[str]
    failed_uri: Optional[str]


def read_data(path: Union[Path, str]) -> Union[bytes, str]:
    """
    helper function to read data (file or network)
    """

    LOGGER.debug(f'Attempting to read {path}')

    if isinstance(path, Path) or not path.startswith(('http', 's3')):
        LOGGER.debug('local file on disk')
        with Path(path).open('rb') as fh:
            return fh.read()
    else:
        LOGGER.debug('network file')
        with urlopen(path) as r:
            return r.read()


def url_join(*parts: str) -> str:
    """
    helper function to join a URL from a number of parts/fragments.
    Implemented because urllib.parse.urljoin strips subpaths from
    host urls if they are specified

    Per https://github.com/geopython/pygeoapi/issues/695

    :param parts: list of parts to join

    :returns: str of resulting URL
    """

    return '/'.join([p.strip().strip('/') for p in parts]).rstrip('/')


def get_envelope(coords_list: List[List[float]]) -> list:
    """
    helper function to get the envelope for a given coordinates
    list through the Shapely API.

    :param coords_list: list of coordinates

    :returns: list of the envelope's coordinates
    """

    coords = [tuple(item) for item in coords_list]
    polygon = Polygon(coords)
    bounds = polygon.bounds
    return [[bounds[0], bounds[3]],
            [bounds[2], bounds[1]]]


class UrlPrefetcher:
    """ Prefetcher to get HTTP headers for specific URLs.
    Allows a maximum of 1 redirect by default.
    """

    def __init__(self):
        self._session = Session()
        self._session.max_redirects = 1

    def get_headers(self, url: str, **kwargs) -> CaseInsensitiveDict:
        """ Issues an HTTP HEAD request to the given URL.
        Returns a case-insensitive dictionary of all headers.
        If the request times out (defaults to 1 second unless `timeout`
        keyword argument is set), or the response has a bad status code,
        an empty dictionary is returned.
        """
        kwargs.setdefault('timeout', 1)
        kwargs.setdefault('allow_redirects', True)
        try:
            response = self._session.head(url, **kwargs)
            response.raise_for_status()
        except Exception:  # noqa
            return CaseInsensitiveDict()
        return response.headers


def bbox2geojsongeometry(bbox: list) -> dict:
    """
    Converts bbox values into GeoJSON geometry

    :param bbox: `list` of minx, miny, maxx, maxy

    :returns: `dict` of GeoJSON geometry
    """

    b = box(*bbox, ccw=False)
    return geom_to_geojson(b)


def get_from_headers(headers: dict, header_name: str) -> str:
    """
    Gets case insensitive value from dictionary.
    This is particularly useful when trying to get
    headers from Starlette and Flask without issue

    :param headers: `dict` of request headers.
    :param header_name: Name of request header.

    :returns: `str` value of header
    """

    cleaned_headers = {k.strip().lower(): v for k, v in headers.items()}
    return cleaned_headers.get(header_name.lower(), '')


def get_choice_from_headers(headers: dict,
                            header_name: str,
                            all: bool = False) -> Union[str, List[str]]:
    """
    Gets choices from a request dictionary,
    considering numerical ordering of preferences.
    Supported are complex preference strings (e.g. "fr-CH, fr;q=0.9, en;q=0.8")

    :param headers: `dict` of request headers.
    :param header_name: Name of request header.
    :param all: bool to return one or all header values.

    :returns: Sorted choice or choices from header
    """

    # Select header of interest
    header = get_from_headers(headers=headers, header_name=header_name)
    if header == '':
        return

    # Parse choices, extracting optional q values (defaults to 1.0)
    choices = []
    for i, part in enumerate(header.split(',')):
        match = re.match(r'^([^;]+)(?:;q=([\d.]+))?$', part.strip())
        if match:
            value, q_value = match.groups()
            q_value = float(q_value) if q_value else 1.0

            # Sort choices by q value and index
            if 0 <= q_value <= 1:
                heappush(choices, (1 / q_value, i, value))

    # Drop q value
    sorted_choices = [choice[-1] for choice in choices]

    # Return one or all choices
    return sorted_choices if all else sorted_choices[0]


def get_dataset_formatters(dataset: dict) -> dict:
    """
    Helper function to derive all formatters for a collection

    :param dataset: `dict` of dataset resource definition

    :returns: `dict` of formatters
    """

    dataset_formatters = {}
    provider_type = get_provider_default(dataset['providers'])['type']

    for key, value in PLUGINS['formatter'].items():
        # workaround to keep items-based collections supporting CSV
        if provider_type not in ['feature', 'record']:
            continue

        df2 = load_plugin('formatter', {'name': key})
        dataset_formatters[key] = df2

    for df in dataset.get('formatters', []):
        df2 = load_plugin('formatter', df)
        dataset_formatters[df2.name] = df2

    return dataset_formatters


def remove_url_auth(url: str) -> str:
    """
    Provide a RFC1738 URL without embedded authentication
    :param url: RFC1738 URL
    :returns: RFC1738 URL without authentication
    """

    u = urlparse(url)
    auth = f'{u.username}:{u.password}@'
    return url.replace(auth, '')


def is_request_allowed(url: str, allow_internal: bool = False) -> bool:
    """
    Test whether an HTTP request is allowed to be executed

    :param url: `str` of URL
    :param allow_internal: `bool` of whether internal requests are
                           allowed (default `False`)

    :returns: `bool` of whether HTTP request execution is allowed
    """

    is_allowed = False

    u = urlparse(url)

    ip = socket.gethostbyname(u.hostname)

    is_private = ipaddress.ip_address(ip).is_private

    if not is_private:
        is_allowed = True
    if is_private and allow_internal:
        is_allowed = True

    return is_allowed



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Martinelli <francesco.martinelli@ingv.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2024 Francesco Martinelli
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
from typing import Any, Tuple, Optional

from pygeoapi.error import GenericError

LOGGER = logging.getLogger(__name__)


class BaseProcessor:
    """generic Processor ABC. Processes are inherited from this class"""

    def __init__(self, processor_def: dict, process_metadata: dict):
        """
        Initialize object

        :param processor_def: processor definition
        :param process_metadata: process metadata `dict`

        :returns: pygeoapi.processor.base.BaseProvider
        """

        self.name = processor_def['name']
        self.metadata = process_metadata
        self.supports_outputs = False
        self.allow_internal_requests = processor_def.get(
            'allow_internal_requests', False)

    def set_job_id(self, job_id: str) -> None:
        """
        Set the job_id within the processor
        To be implemented by derived classes where required.

        :param job_id: the job_id assigned to the request by the Manager.
                       The function should be called by the Manager upon
                       assigning the job_id. The job_id is intended to be used
                       by derived classes, e.g. to write temporary files where
                       filenames contains the string job_id.

        :returns: `None`
        """

        pass

    def execute(self, data: dict, outputs: Optional[dict] = None
                ) -> Tuple[str, Any]:
        """
        execute the process

        :param data: Dict with the input data that the process needs in order
                     to execute
        :param outputs: `dict` or `list` to optionally specify the subset of
                        required outputs - defaults to all outputs.
                        The value of any key may be an object and include the
                        property `transmissionMode` - defaults to `value`.
        :returns: tuple of MIME type and process response
                  (string, bytes, list or dict)
        """

        raise NotImplementedError()

    def __repr__(self):
        return f'<BaseProcessor> {self.name}'


class ProcessorGenericError(GenericError):
    """processor generic error"""
    pass


class ProcessorExecuteError(ProcessorGenericError):
    """query / backend error"""
    default_msg = "generic error (check logs)"


class JobError(Exception):
    pass


class JobNotFoundError(JobError):
    pass


class JobResultNotFoundError(JobError):
    pass


class ProcessError(Exception):
    pass


class UnknownProcessError(ProcessError):
    pass



# =================================================================
#
# Authors: Joana Simoes <jo@byteroad.net>
#
# Copyright (c) 2023 Joana Simoes
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
import requests
from urllib.parse import urlparse

from pygeoapi.provider.base_mvt import BaseMVTProvider
from pygeoapi.provider.base import (ProviderConnectionError,
                                    ProviderGenericError,
                                    ProviderInvalidQueryError)
from pygeoapi.provider.tile import ProviderTileNotFoundError
from pygeoapi.models.provider.base import (
    TileSetMetadata, TileMatrixSetEnum, LinkType)
from pygeoapi.util import is_url, url_join

LOGGER = logging.getLogger(__name__)


class MVTElasticProvider(BaseMVTProvider):
    """MVT Elastic Provider
    Provider for serving tiles rendered with the Elasticsearch
    Vector Tile API
    https://www.elastic.co/guide/en/elasticsearch/reference/current/search-vector-tile-api.html
    As of 12/23, elastic does not provide any tileset metadata.
    """

    def __init__(self, BaseMVTProvider):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.MVT.MVTElasticProvider
        """

        super().__init__(BaseMVTProvider)

        if not is_url(self.data):
            msg = 'Wrong input format for Elasticsearch MVT'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        url = urlparse(self.data)
        baseurl = f'{url.scheme}://{url.netloc}'
        param_type = '?f=mvt'
        layer = f'/{self.get_layer()}'

        LOGGER.debug('Extracting layer name from URL')
        LOGGER.debug(f'Layer: {layer}')

        tilepath = f'{layer}/tiles'
        servicepath = f'{tilepath}/{{tileMatrixSetId}}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}{param_type}'  # noqa

        self._service_url = url_join(baseurl, servicepath)

        self._service_metadata_url = url_join(
            self.service_url.split('{tileMatrix}/{tileRow}/{tileCol}')[0],
            'metadata')

    def __repr__(self):
        return f'<MVTElasticProvider> {self.data}'

    @property
    def service_url(self):
        return self._service_url

    @property
    def service_metadata_url(self):
        return self._service_metadata_url

    def get_layer(self):
        """
        Extracts layer name from url

        :returns: layer name
        """

        if not is_url(self.data):
            msg = 'Wrong input format for Elasticsearch MVT'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        url = urlparse(self.data)

        if ('/{z}/{x}/{y}' not in url.path):
            msg = 'Wrong input format for Elasticsearch MVT'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        layer = url.path.split('/{z}/{x}/{y}')[0]

        LOGGER.debug(layer)
        LOGGER.debug('Removing leading "/"')
        return layer[1:]

    def get_tiling_schemes(self):

        "Only WebMercatorQuad tiling scheme is supported in Elasticsearch"
        return [
                TileMatrixSetEnum.WEBMERCATORQUAD.value
            ]

    def get_tiles_service(self, baseurl=None, servicepath=None,
                          dirpath=None, tile_type=None):
        """
        Gets mvt service description

        :param baseurl: base URL of endpoint
        :param servicepath: base path of URL
        :param dirpath: directory basepath (equivalent of URL)
        :param tile_type: tile format type

        :returns: `dict` of item tile service
        """

        super().get_tiles_service(baseurl, servicepath,
                                  dirpath, tile_type)

        self._service_url = servicepath
        return self.get_tms_links()

    def get_tiles(self, layer=None, tileset=None,
                  z=None, y=None, x=None, format_=None):
        """
        Gets tile

        :param layer: mvt tile layer
        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an encoded mvt tile
        """
        if format_ == 'mvt':
            format_ = self.format_type

        if is_url(self.data):
            url = urlparse(self.data)
            base_url = f'{url.scheme}://{url.netloc}'

            if url.query:
                url_query = f'?{url.query}'
            else:
                url_query = ''

            try:
                with requests.Session() as session:
                    data = {'fields': ['*']}
                    session.get(base_url)
                    resp = session.get(f'{base_url}/{layer}/{z}/{x}/{y}{url_query}', json=data)  # noqa

                    if resp.status_code == 404:
                        if (self.is_in_limits(TileMatrixSetEnum.WEBMERCATORQUAD.value, z, x, y)): # noqa
                            return None
                        raise ProviderTileNotFoundError

                    resp.raise_for_status()
                    return resp.content
            except requests.exceptions.RequestException as e:
                LOGGER.debug(e)
                if resp.status_code < 500:
                    raise ProviderInvalidQueryError  # Client is sending an invalid request # noqa
                raise ProviderGenericError  # Server error
        else:
            msg = 'Wrong input format for Elasticsearch MVT'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

    def get_html_metadata(self, dataset, server_url, layer, tileset,
                          title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt')  # noqa
        metadata_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/metadata')

        metadata = dict()
        metadata['id'] = dataset
        metadata['title'] = title
        metadata['tileset'] = tileset
        metadata['collections_path'] = service_url
        metadata['json_url'] = f'{metadata_url}?f=json'

        return metadata

    def get_default_metadata(self, dataset, server_url, layer, tileset,
                             title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt')  # noqa

        content = {}
        tiling_schemes = self.get_tiling_schemes()
        # Default values
        tileMatrixSetURI = tiling_schemes[0].tileMatrixSetURI
        crs = tiling_schemes[0].crs
        # Checking the selected matrix in configured tiling_schemes
        for schema in tiling_schemes:
            if (schema.tileMatrixSet == tileset):
                crs = schema.crs
                tileMatrixSetURI = schema.tileMatrixSetURI

                tiling_scheme_url = url_join(
                    server_url, f'/TileMatrixSets/{schema.tileMatrixSet}')
                tiling_scheme_url_type = "application/json"
                tiling_scheme_url_title = f'{schema.tileMatrixSet} tile matrix set definition' # noqa

                tiling_scheme = LinkType(href=tiling_scheme_url,
                                         rel="http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme", # noqa
                                         type_=tiling_scheme_url_type,
                                         title=tiling_scheme_url_title)

        if tiling_scheme is None:
            msg = f'Could not identify a valid tiling schema'  # noqa
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        content = TileSetMetadata(title=title, description=description,
                                  keywords=keywords, crs=crs,
                                  tileMatrixSetURI=tileMatrixSetURI)

        links = []
        service_url_link_type = "application/vnd.mapbox-vector-tile"
        service_url_link_title = f'{tileset} vector tiles for {layer}'
        service_url_link = LinkType(href=service_url, rel="item",
                                    type_=service_url_link_type,
                                    title=service_url_link_title)

        links.append(tiling_scheme)
        links.append(service_url_link)

        content.links = links

        return content.model_dump(exclude_none=True)



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================


from contextlib import contextmanager
import json
import logging
from pathlib import Path
from typing import Any, Tuple

import tinydb
from filelock import FileLock

from pygeoapi.formats import FORMAT_TYPES, F_JSON, F_JSONLD
from pygeoapi.process.base import (
    JobNotFoundError,
    JobResultNotFoundError,
)
from pygeoapi.process.manager.base import BaseManager
from pygeoapi.util import JobStatus

LOGGER = logging.getLogger(__name__)


class TinyDBManager(BaseManager):
    """TinyDB Manager"""

    def __init__(self, manager_def: dict):
        """
        Initialize object

        :param manager_def: manager definition

        :returns: `pygeoapi.process.manager.base.BaseManager`
        """

        super().__init__(manager_def)
        self.is_async = True
        self.supports_subscribing = True

    @contextmanager
    def _db(self):
        self.lock = FileLock(f"{self.connection}.lock")
        with self.lock:
            with tinydb.TinyDB(self.connection) as db:
                yield db

    def destroy(self) -> bool:
        """
        Destroy manager

        :returns: `bool` status of result
        """

        with self._db as db:
            db.purge()

        return True

    def get_jobs(self, status: JobStatus = None, limit=None, offset=None
                 ) -> dict:
        """
        Get jobs

        :param status: job status (accepted, running, successful,
                       failed, results) (default is all)
        :param limit: number of jobs to return
        :param offset: pagination offset

        :returns: dict of list of jobs (identifier, status, process identifier)
                  and numberMatched
        """

        with self._db() as db:
            jobs_list = db.all()

        number_matched = len(jobs_list)

        if offset:
            jobs_list = jobs_list[offset:]

        if limit:
            jobs_list = jobs_list[:limit]

        return {
            'jobs': jobs_list,
            'numberMatched': number_matched
        }

    def add_job(self, job_metadata: dict) -> str:
        """
        Add a job

        :param job_metadata: `dict` of job metadata

        :returns: identifier of added job
        """

        with self._db() as db:
            doc_id = db.insert(job_metadata)

        return doc_id  # noqa

    def update_job(self, job_id: str, update_dict: dict) -> bool:
        """
        Updates a job

        :param job_id: job identifier
        :param update_dict: `dict` of property updates

        :returns: `bool` of status result
        """

        with self._db() as db:
            db.update(update_dict, tinydb.where('identifier') == job_id)

        return True

    def delete_job(self, job_id: str) -> bool:
        """
        Deletes a job

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :return `bool` of status result
        """
        # delete result file if present
        job_result = self.get_job(job_id)
        location = job_result.get('location')
        if location and self.output_dir is not None:
            Path(location).unlink()

        with self._db() as db:
            removed = bool(db.remove(tinydb.where('identifier') == job_id))

        return removed

    def get_job(self, job_id: str) -> dict:
        """
        Get a single job

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :returns: `dict`  # `pygeoapi.process.manager.Job`
        """

        query = tinydb.Query()
        with self._db() as db:
            found = db.search(query.identifier == job_id)
        if found is not None:
            try:
                return found[0]
            except IndexError:
                raise JobNotFoundError()
        else:
            raise JobNotFoundError()

    def get_job_result(self, job_id: str) -> Tuple[str, Any]:
        """
        Get a job's status, and actual output of executing the process

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :raises JobResultNotFoundError: if the job-related result cannot
                                        be returned
        :returns: `tuple` of mimetype and raw output
        """

        job_result = self.get_job(job_id)
        location = job_result.get('location')
        mimetype = job_result.get('mimetype')
        job_status = JobStatus[job_result['status']]

        if not job_status == JobStatus.successful:
            # Job is incomplete
            return (None,)
        if not location:
            LOGGER.warning(f'job {job_id!r} -  unknown result location')
            raise JobResultNotFoundError()
        else:
            try:
                location = Path(location)
                if mimetype in (None, FORMAT_TYPES[F_JSON],
                                FORMAT_TYPES[F_JSONLD]):
                    with location.open('r', encoding='utf-8') as filehandler:
                        result = json.load(filehandler)
                else:
                    with location.open('rb') as filehandler:
                        result = filehandler.read()
            except (TypeError, FileNotFoundError, json.JSONDecodeError):
                raise JobResultNotFoundError()
            else:
                return mimetype, result

    def __repr__(self):
        return f'<TinyDBManager> {self.name}'



# =================================================================

# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
# Copyright (c) 2024 Bernhard Mallinger
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================


from copy import deepcopy
from http import HTTPStatus
import logging
from typing import Tuple

from pygeoapi.crs import transform_bbox
from pygeoapi.formats import F_JSON, FORMAT_TYPES
from pygeoapi.openapi import get_oas_30_parameters
from pygeoapi.plugin import load_plugin
from pygeoapi.provider import filter_providers_by_type, get_provider_by_type
from pygeoapi.provider.base import (
    ProviderGenericError, ProviderInvalidDataError
)
from pygeoapi.util import to_json, filter_dict_by_key_value

from . import APIRequest, API, validate_datetime, validate_subset

LOGGER = logging.getLogger(__name__)

CONFORMANCE_CLASSES = [
    'http://www.opengis.net/spec/ogcapi-maps-1/1.0/conf/core'
]


DEFAULT_CRS = 'http://www.opengis.net/def/crs/EPSG/0/4326'


def get_collection_map(api: API, request: APIRequest,
                       dataset: str, style: str | None = None
                       ) -> Tuple[dict, int, str]:
    """
    Returns an image of a collection map

    :param request: A request object
    :param dataset: dataset name
    :param style: style name

    :returns: tuple of headers, status code, content
    """

    query_args = {}

    format_ = request.format or 'png'
    headers = request.get_response_headers(**api.api_headers)
    LOGGER.debug('Processing query parameters')

    LOGGER.debug('Loading provider')
    try:
        collection_def = get_provider_by_type(
            api.config['resources'][dataset]['providers'], 'map')

        p = load_plugin('provider', collection_def)
    except KeyError:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'collection does not exist'
        }
        headers['Content-type'] = 'application/json'
        LOGGER.error(exception)
        return headers, HTTPStatus.NOT_FOUND, to_json(
            exception, api.pretty_print)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    query_args['format_'] = request.params.get('f', 'png')
    query_args['style'] = style
    query_args['crs'] = collection_def.get('crs', DEFAULT_CRS)
    query_args['bbox_crs'] = request.params.get(
        'bbox-crs', DEFAULT_CRS
    )
    query_args['transparent'] = request.params.get('transparent', True)

    try:
        query_args['width'] = int(request.params.get('width', 500))
        query_args['height'] = int(request.params.get('height', 300))
    except ValueError:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'invalid width/height'
        }
        headers['Content-type'] = 'application/json'
        LOGGER.error(exception)
        return headers, HTTPStatus.BAD_REQUEST, to_json(
            exception, api.pretty_print)

    LOGGER.debug('Processing bbox parameter')
    try:
        bbox = request.params.get('bbox').split(',')
        if len(bbox) != 4:
            exception = {
                'code': 'InvalidParameterValue',
                'description': 'bbox values should be minx,miny,maxx,maxy'
            }
            headers['Content-type'] = 'application/json'
            LOGGER.error(exception)
            return headers, HTTPStatus.BAD_REQUEST, to_json(
                exception, api.pretty_print)
    except AttributeError:
        bbox = api.config['resources'][dataset]['extents']['spatial']['bbox']  # noqa
    try:
        bbox = [float(c) for c in bbox]
    except ValueError:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'bbox values must be numbers'
        }
        headers['Content-type'] = 'application/json'
        LOGGER.error(exception)
        return headers, HTTPStatus.BAD_REQUEST, to_json(
            exception, api.pretty_print)

    if query_args['bbox_crs'] != query_args['crs']:
        LOGGER.debug(f'Reprojecting bbox CRS: {query_args["crs"]}')
        bbox = transform_bbox(bbox, query_args['bbox_crs'], query_args['crs'])

    query_args['bbox'] = bbox

    LOGGER.debug('Processing datetime parameter')
    datetime_ = request.params.get('datetime')
    try:
        query_args['datetime_'] = validate_datetime(
            api.config['resources'][dataset]['extents'], datetime_)
    except ValueError as err:
        msg = str(err)
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    if 'subset' in request.params:
        # TODO get subsets from provider
        subsets = deepcopy(api.config['resources'][dataset]['extents'])
        subsets.pop('spatial', None)  # bbox
        subsets.pop('temporal', None)  # datetime
        LOGGER.debug('Processing subset parameter')
        try:
            query_args['subsets'] = validate_subset(
                request.params['subset'] or '')
        except (AttributeError, ValueError) as err:
            msg = f'Invalid subset: {err}'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, format_,
                'InvalidParameterValue', msg)

        for sk in query_args['subsets'].keys():
            if sk not in subsets.keys():
                msg = f'Subset not found; valid values are {subsets}'
                return api.get_exception(
                    HTTPStatus.BAD_REQUEST, headers, format_,
                    'InvalidParameterValue', msg)

    if request.params.get('properties'):
        try:
            fields = p.get_fields() or {}
        except NotImplementedError:
            msg = 'No properties implemented'
            headers['Content-Type'] = FORMAT_TYPES[F_JSON]
            return api.get_exception(
                HTTPStatus.NOT_IMPLEMENTED, headers, format_,
                'InvalidParameterValue', msg)

        LOGGER.debug('Processing properties parameter')
        properties = request.params.get('properties') or []
        if isinstance(properties, str):
            properties = properties.split(',')

        if properties and not any((fld in properties)
                                  for fld in fields.keys()):
            msg = f'Invalid property; valid property names are {list(fields.keys())}'  # noqa
            headers['Content-Type'] = FORMAT_TYPES[F_JSON]
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)

        query_args['select_properties'] = properties

    LOGGER.debug('Generating map')
    try:
        data = p.query(**query_args)
    except (ProviderGenericError, ProviderInvalidDataError) as err:
        headers['Content-Type'] = FORMAT_TYPES[F_JSON]
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    mt = collection_def['format']['name']

    if format_ == mt:
        headers['Content-Type'] = collection_def['format']['mimetype']
        return headers, HTTPStatus.OK, data
    elif format_ in [None, 'html']:
        headers['Content-Type'] = collection_def['format']['mimetype']
        return headers, HTTPStatus.OK, data
    else:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'invalid format parameter'
        }
        LOGGER.error(exception)
        return headers, HTTPStatus.BAD_REQUEST, to_json(
            data, api.pretty_print)


def get_collection_map_legend(api: API, request: APIRequest,
                              dataset: str, style: str | None = None
                              ) -> Tuple[dict, int, str]:
    """
    Returns an image of a collection map legend

    :param request: A request object
    :param dataset: dataset name
    :param style: style name

    :returns: tuple of headers, status code, content
    """

    format_ = 'png'
    headers = request.get_response_headers(**api.api_headers)
    LOGGER.debug('Processing query parameters')

    LOGGER.debug('Loading provider')
    try:
        collection_def = get_provider_by_type(
            api.config['resources'][dataset]['providers'], 'map')

        p = load_plugin('provider', collection_def)
    except KeyError:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'collection does not exist'
        }
        LOGGER.error(exception)
        return headers, HTTPStatus.NOT_FOUND, to_json(
            exception, api.pretty_print)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    LOGGER.debug('Generating legend')
    try:
        data = p.get_legend(style, request.params.get('f', 'png'))
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    mt = collection_def['format']['name']

    if format_ == mt:
        headers['Content-Type'] = collection_def['format']['mimetype']
        return headers, HTTPStatus.OK, data
    else:
        exception = {
            'code': 'InvalidParameterValue',
            'description': 'invalid format parameter'
        }
        LOGGER.error(exception)
        return headers, HTTPStatus.BAD_REQUEST, to_json(
            data, api.pretty_print)


def get_oas_30(cfg: dict, locale: str) -> tuple[list[dict[str, str]], dict[str, dict]]:  # noqa
    """
    Get OpenAPI fragments

    :param cfg: `dict` of configuration
    :param locale: `str` of locale

    :returns: `tuple` of `list` of tag objects, and `dict` of path objects
    """

    from pygeoapi.openapi import OPENAPI_YAML, get_visible_collections

    LOGGER.debug('setting up maps endpoints')

    paths = {}

    collections = filter_dict_by_key_value(cfg['resources'],
                                           'type', 'collection')

    parameters = get_oas_30_parameters(cfg, locale)
    for k, v in get_visible_collections(cfg).items():
        map_extension = filter_providers_by_type(
            collections[k]['providers'], 'map')

        if map_extension:
            mp = load_plugin('provider', map_extension)

            coll_properties = deepcopy(parameters)['properties']
            coll_properties['schema']['items']['enum'] = list(mp.fields.keys())

            map_f = deepcopy(parameters['f'])
            map_f['schema']['enum'] = [map_extension['format']['name']]
            map_f['schema']['default'] = map_extension['format']['name']

            pth = f'/collections/{k}/map'
            paths[pth] = {
                'get': {
                    'summary': 'Get map',
                    'description': f"{v['description']} map",
                    'tags': [k],
                    'operationId': 'getMap',
                    'parameters': [
                        {'$ref': '#/components/parameters/bbox'},
                        {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"},  # noqa
                        {'$ref': f"{OPENAPI_YAML['oamaps']}#/components/parameters/subset"},  # noqa
                        {
                            'name': 'width',
                            'in': 'query',
                            'description': 'Response image width',
                            'required': False,
                            'schema': {
                                'type': 'integer',
                            },
                            'style': 'form',
                            'explode': False
                        },
                        {
                            'name': 'height',
                            'in': 'query',
                            'description': 'Response image height',
                            'required': False,
                            'schema': {
                                'type': 'integer',
                            },
                            'style': 'form',
                            'explode': False
                        },
                        {
                            'name': 'transparent',
                            'in': 'query',
                            'description': 'Background transparency of map (default=true).',  # noqa
                            'required': False,
                            'schema': {
                                'type': 'boolean',
                                'default': True,
                            },
                            'style': 'form',
                            'explode': False
                        },
                        {'$ref': '#/components/parameters/bbox-crs-epsg'},
                        map_f
                    ],
                    'responses': {
                        '200': {
                            'description': 'Response',
                            'content': {
                                'application/json': {}
                            }
                        },
                        '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                        '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"},  # noqa
                    }
                }
            }
            if coll_properties['schema']['items']['enum']:
                paths[pth]['get']['parameters'].append(coll_properties)

            if mp.time_field is not None:
                paths[pth]['get']['parameters'].append(
                    {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"})  # noqa

    return [{'name': 'maps'}], {'paths': paths}



# =================================================================
#
# Authors: Joana Simoes <jo@byteroad.net>
#          Francesco Bartoli <xbartolone@gmail.com>
#
# Copyright (c) 2023 Joana Simoes
# Copyright (c) 2025 Francesco Bartoli
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import json
import logging
import requests
from pathlib import Path
from urllib.parse import urlparse

from pygeoapi.provider.tile import (
    ProviderTileNotFoundError)
from pygeoapi.provider.base import (ProviderConnectionError,
                                    ProviderInvalidQueryError,
                                    ProviderGenericError)
from pygeoapi.provider.base_mvt import BaseMVTProvider
from pygeoapi.models.provider.base import (
    TileSetMetadata, TileMatrixSetEnum, LinkType)
from pygeoapi.models.provider.mvt import MVTTilesJson

from pygeoapi.util import is_url, url_join

LOGGER = logging.getLogger(__name__)


class MVTTippecanoeProvider(BaseMVTProvider):
    """MVT Tippecanoe Provider
    Provider for serving tiles generated with Mapbox Tippecanoe
    https://github.com/mapbox/tippecanoe
    It supports both, tiles from a an url or a path on disk.
    Tippecanoe also provides a TileSet Metadata in a file called
    "metadata.json".
    """

    def __init__(self, BaseMVTProvider):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.MVT.MVTTippecanoeProvider
        """

        super().__init__(BaseMVTProvider)

        # Pre-rendered tiles served from a static url
        if is_url(self.data):
            url = urlparse(self.data)
            baseurl = f'{url.scheme}://{url.netloc}'
            layer = f'/{self.get_layer()}'

            LOGGER.debug('Extracting layer name from URL')
            LOGGER.debug(f'Layer: {layer}')

            tilepath = f'{layer}/tiles'
            servicepath = f'{tilepath}/{{tileMatrixSetId}}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt'  # noqa

            self._service_url = url_join(baseurl, servicepath)

            self._service_metadata_url = url_join(
                self.service_url.split('{tileMatrix}/{tileRow}/{tileCol}')[0],
                'metadata')

            metadata_path = f'{baseurl}/{layer}/metadata.json'
            head = requests.head(metadata_path)
            if head.status_code != 200:
                msg = f'Service metadata does not exist: {metadata_path}'
                LOGGER.error(msg)
                LOGGER.warning(msg)
            self._service_metadata_url = metadata_path

        # Pre-rendered tiles served from a local path
        else:
            data_path = Path(self.data)
            if not data_path.exists():
                msg = f'Service does not exist: {self.data}'
                LOGGER.error(msg)
                raise ProviderConnectionError(msg)
            self._service_url = data_path
            metadata_path = data_path.joinpath('metadata.json')
            if not metadata_path.exists():
                msg = f'Service metadata does not exist: {metadata_path.name}'
                LOGGER.error(msg)
                LOGGER.warning(msg)
            self._service_metadata_url = metadata_path

    def __repr__(self):
        return f'<MVTTippecanoeProvider> {self.data}'

    def get_layer(self):
        """
        Extracts layer name from url or data path

        :returns: layer name
        """

        if is_url(self.data):
            url = urlparse(self.data)

            if ('/{z}/{x}/{y}' not in url.path):
                msg = f'This url template is not supported yet: {url.path}'
                LOGGER.error(msg)
                raise ProviderConnectionError(msg)

            layer = url.path.split('/{z}/{x}/{y}')[0]

            LOGGER.debug(layer)
            LOGGER.debug('Removing leading "/"')
            return layer[1:]

        else:
            return Path(self.data).name

    def get_tiling_schemes(self):
        "Only WebMercatorQuad tiling scheme is supported in tippecanoe"

        return [TileMatrixSetEnum.WEBMERCATORQUAD.value]

    def get_tiles_service(self, baseurl=None, servicepath=None,
                          dirpath=None, tile_type=None):
        """
        Gets mvt service description

        :param baseurl: base URL of endpoint
        :param servicepath: base path of URL
        :param dirpath: directory basepath (equivalent of URL)
        :param tile_type: tile format type

        :returns: `dict` of item tile service
        """

        super().get_tiles_service(baseurl, servicepath,
                                  dirpath, tile_type)

        self._service_url = servicepath
        return self.get_tms_links()

    def get_tiles_from_url(self, layer=None, tileset=None,
                           z=None, y=None, x=None, format_=None):
        """
        Gets tile from a static url (e.g.: bucket, file server)

        :param layer: mvt tile layer
        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an encoded mvt tile
        """

        url = urlparse(self.data)
        base_url = f'{url.scheme}://{url.netloc}'

        extension = Path(url.path).suffix  # e.g. ".pbf"

        try:
            with requests.Session() as session:
                session.get(base_url)
                resp = session.get(f'{base_url}/{layer}/{z}/{x}/{y}{extension}')  # noqa

                if resp.status_code == 404:
                    if (self.is_in_limits(TileMatrixSetEnum.WEBMERCATORQUAD.value, z, x, y)): # noqa
                        return None
                    raise ProviderTileNotFoundError

                resp.raise_for_status()
                return resp.content
        except requests.exceptions.RequestException as e:
            LOGGER.debug(e)
            if resp.status_code < 500:
                raise ProviderInvalidQueryError  # Client is sending an invalid request # noqa
            raise ProviderGenericError  # Server error

    def get_tiles_from_disk(self, layer=None, tileset=None,
                            z=None, y=None, x=None, format_=None):
        """
        Gets tile from a path on disk

        :param layer: mvt tile layer
        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an encoded mvt tile
        """

        try:
            service_url_path = self.service_url.joinpath(f'{z}/{x}/{y}.{format_}')  # noqa
            with open(service_url_path, mode='rb') as tile:
                return tile.read()
        except FileNotFoundError as err:
            if (self.is_in_limits(TileMatrixSetEnum.WEBMERCATORQUAD.value, z, x, y)): # noqa
                return None
            raise ProviderTileNotFoundError(err)

    def get_tiles(self, layer=None, tileset=None,
                  z=None, y=None, x=None, format_=None):
        """
        Gets tile

        :param layer: mvt tile layer
        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an encoded mvt tile
        """

        if format_ == 'mvt':
            format_ = self.format_type

        if isinstance(self.service_url, Path):
            return self.get_tiles_from_disk(layer, tileset, z, y, x, format_)
        elif is_url(self.data):
            return self.get_tiles_from_url(layer, tileset, z, y, x, format_)
        else:
            msg = 'Wrong input format for Tippecanoe MVT'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

    def get_metadata_from_URL(self, service_metadata_url=None):
        """
        Gets the metadata JSON from disk or URL

        :param service_metadata_url: local path or URL

        :returns: the metadata JSON
        """
        if (not service_metadata_url):
            service_metadata_url = self.service_metadata_url

        metadata_json_content = ''
        if isinstance(service_metadata_url, Path):
            if service_metadata_url.exists():
                with open(service_metadata_url, 'r') as md_file:
                    metadata_json_content = json.loads(md_file.read())
        elif is_url(service_metadata_url):
            with requests.Session() as session:
                resp = session.get(service_metadata_url)
                resp.raise_for_status()
                metadata_json_content = json.loads(resp.content)
        else:
            msg = f'Wrong data path configuration: {service_metadata_url}'  # noqa
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        return metadata_json_content

    def get_html_metadata(self, dataset, server_url, layer, tileset,
                          title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt')  # noqa
        metadata_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/metadata')

        metadata = dict()
        metadata['id'] = dataset
        metadata['title'] = title
        metadata['tileset'] = tileset
        metadata['collections_path'] = service_url
        metadata['json_url'] = f'{metadata_url}?f=json'

        try:
            metadata_json_content = self.get_metadata_from_URL(self.service_metadata_url)  # noqa

            content = MVTTilesJson(**metadata_json_content)
            content.tiles = service_url
            content.vector_layers = json.loads(
                metadata_json_content["json"])["vector_layers"]
            metadata['metadata'] = content.model_dump()
            # Some providers may not implement tilejson metadata
            metadata['tilejson_url'] = f'{metadata_url}?f=tilejson'
        except ProviderConnectionError:
            # No vendor metadata JSON
            pass

        return metadata

    def get_default_metadata(self, dataset, server_url, layer, tileset,
                             title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt')  # noqa

        content = {}
        tiling_schemes = self.get_tiling_schemes()
        # Default values
        tileMatrixSetURI = tiling_schemes[0].tileMatrixSetURI
        crs = tiling_schemes[0].crs

        tiling_scheme = None

        # Checking the selected matrix in configured tiling_schemes
        for schema in tiling_schemes:
            if (schema.tileMatrixSet == tileset):
                crs = schema.crs
                tileMatrixSetURI = schema.tileMatrixSetURI

                tiling_scheme_url = url_join(
                    server_url, f'/TileMatrixSets/{schema.tileMatrixSet}')
                tiling_scheme_url_type = "application/json"
                tiling_scheme_url_title = f'{schema.tileMatrixSet} tile matrix set definition'  # noqa

                tiling_scheme = LinkType(href=tiling_scheme_url,
                                         rel="http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",  # noqa
                                         type_=tiling_scheme_url_type,
                                         title=tiling_scheme_url_title)

        if tiling_scheme is None:
            msg = "Could not identify a valid tiling schema"
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        content = TileSetMetadata(title=title, description=description,
                                  keywords=keywords, crs=crs,
                                  tileMatrixSetURI=tileMatrixSetURI)

        links = []

        service_url_link_type = "application/vnd.mapbox-vector-tile"
        service_url_link_title = f'{tileset} vector tiles for {layer}'
        service_url_link = LinkType(href=service_url, rel="item",
                                    type_=service_url_link_type,
                                    title=service_url_link_title)

        links.append(tiling_scheme)
        links.append(service_url_link)

        content.links = links

        return content.model_dump(exclude_none=True)

    def get_vendor_metadata(self, dataset, server_url, layer, tileset,
                            title, description, keywords, **kwargs):
        """
        Gets tile metadata in tilejson format
        """

        try:
            metadata_json_content = self.get_metadata_from_URL(self.service_metadata_url)  # noqa

            service_url = url_join(
                server_url,
                f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f=mvt')  # noqa

            content = MVTTilesJson(**metadata_json_content)
            content.tiles = service_url
            content.vector_layers = json.loads(
                metadata_json_content["json"])["vector_layers"]
            return content.model_dump()
        except ProviderConnectionError:
            msg = f'No tiles metadata json available: {self.service_metadata_url}'  # noqa
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)



# =================================================================
#
# Authors: Benjamin Webb <benjamin.miller.webb@gmail.com>
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2025 Benjamin Webb
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from json.decoder import JSONDecodeError
import logging
from requests import Session
from requests.exceptions import ConnectionError
from urllib.parse import urlparse

from pygeoapi.config import get_config
from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import (
    BaseProvider, ProviderQueryError, ProviderConnectionError,
    ProviderInvalidDataError)
from pygeoapi.util import (
    url_join, get_provider_default, get_base_url, get_typed_value)

LOGGER = logging.getLogger(__name__)

ENTITY = {
    'Thing', 'Things', 'Observation', 'Observations',
    'Location', 'Locations', 'Sensor', 'Sensors',
    'Datastream', 'Datastreams', 'ObservedProperty',
    'ObservedProperties', 'FeatureOfInterest', 'FeaturesOfInterest',
    'HistoricalLocation', 'HistoricalLocations'
}
_EXPAND = {
    'Things': 'Locations,Datastreams',
    'Observations': 'Datastream,FeatureOfInterest',
    'ObservedProperties': 'Datastreams/Thing/Locations',
    'Datastreams': """
        Sensor
        ,ObservedProperty
        ,Thing/Locations
        ,Observations(
            $select=@iot.id;
            $orderby=phenomenonTime_desc
            )
        ,Observations/FeatureOfInterest(
            $select=feature
            )
    """
}
EXPAND = {k: ''.join(v.split()).replace('_', ' ')
          for (k, v) in _EXPAND.items()}


class SensorThingsProvider(BaseProvider):
    """SensorThings API (STA) Provider"""
    expand = EXPAND

    def __init__(self, provider_def):
        """
        STA Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class

        :returns: pygeoapi.provider.sensorthings.SensorThingsProvider
        """
        LOGGER.debug('Setting SensorThings API (STA) provider')
        self.linked_entity = {}
        super().__init__(provider_def)

        self._generate_mappings(provider_def)
        LOGGER.debug(f'STA endpoint: {self.data}, Entity: {self.entity}')

        # Start session
        self.http = Session()
        self.get_fields()

    def get_fields(self):
        """
        Get fields of STA Provider

        :returns: dict of fields
        """
        if not self._fields:
            try:
                r = self._get_response(self._url, {'$top': 1})
                results = r['value'][0]
            except IndexError:
                LOGGER.warning('could not get fields; returning empty set')
                return {}
            except (ConnectionError, ProviderConnectionError):
                msg = f'Unable to contact SensorThings endpoint at {self._url}'
                LOGGER.error(msg)
                raise ProviderConnectionError(msg)

            for (n, v) in results.items():
                if isinstance(v, (int, float)) or \
                   (isinstance(v, (dict, list)) and n in ENTITY):
                    self._fields[n] = {'type': 'number'}
                elif isinstance(v, str):
                    self._fields[n] = {'type': 'string'}

        return self._fields

    @crs_transform
    def query(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None, **kwargs):
        """
        STA query

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)

        :returns: dict of GeoJSON FeatureCollection
        """

        return self._load(offset, limit, resulttype, bbox=bbox,
                          datetime_=datetime_, properties=properties,
                          sortby=sortby, select_properties=select_properties,
                          skip_geometry=skip_geometry)

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        Query STA by id

        :param identifier: feature id

        :returns: dict of single GeoJSON feature
        """
        response = self._get_response(f'{self._url}({identifier})')
        return self._make_feature(response)

    def create(self, item):
        """
        Create a new item

        :param item: `dict` of new item

        :returns: identifier of created item
        """
        response = self.http.post(self._url, json=item)

        if response.status_code == 201:
            location = response.headers.get("Location")
            iotid = location[location.find("(")+1:location.find(")")]

            LOGGER.debug(f'Feature created with @iot.id: {iotid}')
            return get_typed_value(iotid)
        else:
            msg = f"Failed to create item: {response.text}"
            raise ProviderInvalidDataError(msg)

    def update(self, identifier, item):
        """
        Updates an existing item

        :param identifier: feature id
        :param item: `dict` of partial or full item

        :returns: `bool` of update result
        """
        id = f"'{identifier}'" \
             if isinstance(identifier, str) else str(identifier)
        LOGGER.debug(f'Updating @iot.id: {id}')
        response = self.http.put(f"{self._url}({id})", json=item)

        if response.status_code == 200:
            return True
        else:
            msg = f'Failed to update item: {response.text}'
            raise ProviderConnectionError(msg)

    def delete(self, identifier):
        """
        Deletes an existing item

        :param identifier: item id

        :returns: `bool` of deletion result
        """
        id = f"'{identifier}'" \
             if isinstance(identifier, str) else str(identifier)
        LOGGER.debug(f'Deleting @iot.id: {id}')
        response = self.http.delete(f"{self._url}({id})")

        if response.status_code == 200:
            return True
        else:
            msg = f"Failed to delete item: {response.text}"
            raise ProviderConnectionError(msg)

    def _load(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None):
        """
        Private function: Load STA data

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)

        :returns: dict of GeoJSON FeatureCollection
        """

        # Make defaults
        fc = {'type': 'FeatureCollection', 'features': []}
        params = {
            '$skip': str(offset),
            '$top': str(limit)
        }

        if properties or bbox or datetime_:
            params['$filter'] = self._make_filter(properties, bbox, datetime_)

        if sortby:
            params['$orderby'] = self._make_orderby(sortby)

        # Send request
        LOGGER.debug('Sending query')
        if resulttype == 'hits':
            LOGGER.debug('Returning hits')
            params['$count'] = 'true'
            response = self._get_response(url=self._url, params=params)
            fc['numberMatched'] = response.get('@iot.count')
            return fc

        # Make features
        response = self._get_response(url=self._url, params=params)

        matched = response.get('@iot.count')
        if matched:
            fc['numberMatched'] = matched

        # Query if values are less than expected
        v = response.get('value')
        while len(v) < limit:
            try:
                # Ensure we only use provided network location
                next_ = urlparse(response['@iot.nextLink'])._replace(
                    scheme=self.parsed_url.scheme,
                    netloc=self.parsed_url.netloc
                ).geturl()

                LOGGER.debug('Fetching next set of values')
                response = self._get_response(next_)
                v.extend(response['value'])
            except (ProviderConnectionError, KeyError):
                break

        hits_ = min(limit, len(v))
        props = (select_properties, skip_geometry)
        fc['features'] = [self._make_feature(e, *props) for e in v[:hits_]]
        fc['numberReturned'] = hits_

        return fc

    def _make_feature(self, feature, select_properties=[], skip_geometry=False,
                      entity=None):
        """
        Private function: Create feature from entity

        :param feature: `dict` of STA entity
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param entity: SensorThings entity name

        :returns: dict of GeoJSON Feature
        """
        _ = feature[self.id_field]
        id = f"'{_}'" if isinstance(_, str) else str(_)
        f = {
            'type': 'Feature', 'id': id, 'properties': {}, 'geometry': None
        }

        # Make geometry
        if not skip_geometry:
            f['geometry'] = self._geometry(feature, entity)

        # Fill properties block
        try:
            f['properties'] = self._expand_properties(
                feature, select_properties, entity)
        except KeyError as err:
            LOGGER.error(err)
            raise ProviderQueryError(err)

        return f

    def _get_response(self, url, params={}, entity=None, expand=None):
        """
        Private function: Get STA response

        :param url: request url
        :param params: query parameters
        :param entity: SensorThings entity name
        :param expand: SensorThings expand query


        :returns: STA response
        """
        if expand:
            params.update({'$expand': expand})
        else:
            entity_ = entity or self.entity
            params.update({'$expand': self.expand[entity_]})

        r = self.http.get(url, params=params)

        if not r.ok:
            LOGGER.error('Bad http response code')
            raise ProviderConnectionError('Bad http response code')

        try:
            response = r.json()
        except JSONDecodeError as err:
            LOGGER.error('JSON decode error')
            raise ProviderQueryError(err)

        return response

    def _make_filter(self, properties, bbox=[], datetime_=None,
                     entity=None):
        """
        Private function: Make STA filter from query properties

        :param properties: list of tuples (name, value)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param entity: SensorThings entity name

        :returns: STA $filter string of properties
        """
        ret = []
        for (name, value) in properties:
            if name in ENTITY:
                ret.append(f'{name}/@iot.id eq {value}')
            else:
                ret.append(f'{name} eq {value}')

        if bbox:
            entity_ = entity or self.entity
            ret.append(self._make_bbox(bbox, entity_))

        if datetime_ is not None:
            if self.time_field is None:
                msg = 'time_field not enabled for collection'
                LOGGER.error(msg)
                raise ProviderQueryError(msg)

            if '/' in datetime_:
                time_start, time_end = datetime_.split('/')
                if time_start != '..':
                    ret.append(f'{self.time_field} ge {time_start}')
                if time_end != '..':
                    ret.append(f'{self.time_field} le {time_end}')
            else:
                ret.append(f'{self.time_field} eq {datetime_}')

        return ' and '.join(ret)

    @staticmethod
    def _make_bbox(bbox, entity):
        minx, miny, maxx, maxy = bbox
        bbox_ = f'POLYGON(({minx} {miny},{maxx} {miny},{maxx} {maxy},{minx} {maxy},{minx} {miny}))' # noqa
        if entity == 'Things':
            loc = 'Locations/location'
        elif entity == 'Datastreams':
            loc = 'Thing/Locations/location'
        elif entity == 'Observations':
            loc = 'FeatureOfInterest/feature'
        elif entity == 'ObservedProperties':
            loc = 'Datastreams/observedArea'
        return f"st_within({loc},geography'{bbox_}')"

    def _make_orderby(self, sortby):
        """
        Private function: Make STA filter from query properties

        :param sortby: list of dicts (property, order)

        :returns: STA $orderby string
        """
        ret = []
        _map = {'+': 'asc', '-': 'desc'}
        for _ in sortby:
            prop = _['property']
            order = _map[_['order']]
            if prop in ENTITY:
                ret.append(f'{prop}/@iot.id {order}')
            else:
                ret.append(f'{prop} {order}')

        return ','.join(ret)

    def _geometry(self, feature, entity=None):
        """
        Private function: Retrieve STA geometry

        :param feature: SensorThings entity
        :param entity: SensorThings entity name

        :returns: GeoJSON Geometry for feature
        """
        entity_ = entity or self.entity
        try:
            if entity_ == 'Things':
                return feature['Locations'][0]['location']

            elif entity_ == 'Observations':
                return feature['FeatureOfInterest'].pop('feature')

            elif entity_ == 'Datastreams':
                try:
                    return feature['Observations'][0]['FeatureOfInterest'].pop('feature')  # noqa
                except (KeyError, IndexError):
                    return feature['Thing'].pop('Locations')[0]['location']

            elif entity_ == 'ObservedProperties':
                return feature['Datastreams'][0]['Thing']['Locations'][0]['location'] # noqa

        except (KeyError, IndexError):
            LOGGER.warning('No geometry found')
            return None

    def _expand_properties(self, feature, keys=(), uri='',
                           entity=None):
        """
        Private function: Parse STA entity into feature

        :param feature: `dict` of SensorThings entity
        :param keys: keys used in properties block
        :param uri: uri of STA entity
        :param entity: SensorThings entity name

        :returns: dict of SensorThings feature properties
        """
        # Properties filter & display
        keys = (() if not self.properties and not keys else
                set(self.properties) | set(keys))

        entity = entity or self.entity
        if entity == 'Things':
            self._expand_location(feature)
        elif 'Thing' in feature.keys():
            self._expand_location(feature['Thing'])

        # Retain URI if present
        try:
            if self.uri_field is not None:
                uri = feature['properties'][self.uri_field]
        except KeyError:
            msg = f'Unable to find uri field: {self.uri_field}'
            LOGGER.error(msg)
            raise ProviderInvalidDataError(msg)

        # Create intra links
        for k, v in feature.items():
            if k in self.linked_entity:
                feature[k] = [self._get_uri(_v, **self.linked_entity[k])
                              for _v in v]
                LOGGER.debug(f'Created link for {k}')
            elif f'{k}s' in self.linked_entity:
                feature[k] = \
                    self._get_uri(v, **self.linked_entity[f'{k}s'])
                LOGGER.debug(f'Created link for {k}')

        # Make properties block
        if feature.get('properties'):
            feature.update(feature.pop('properties'))

        if keys:
            ret = {k: feature.pop(k) for k in keys}
            feature = ret

        if self.uri_field is not None:
            feature[self.uri_field] = uri

        return feature

    @staticmethod
    def _expand_location(entity):
        """
        Private function: Get STA item uri

        :param entity: `dict` of STA entity

        :returns: None
        """
        try:
            extra_props = entity['Locations'][0]['properties']
            entity['properties'].update(extra_props)
        except (KeyError, IndexError):
            selfLink = entity['@iot.selfLink']
            LOGGER.debug(f'{selfLink} has no Location properties')

    def _get_uri(self, entity, cnm, cid='@iot.id', uri=''):
        """
        Private function: Get STA item uri

        :param entity: `dict` of STA entity
        :param cnm: `str` of OAPI collection name
        :param cid: `str` of OAPI collection id field
        :param uri: `str` of STA entity uri field

        :returns: `str` of item uri
        """
        if uri:
            return entity['properties'][uri]
        else:
            id_ = entity[cid]
            id_ = f"'{id_}'" if isinstance(id_, str) else str(id_)
            uri = (self.rel_link, 'collections', cnm, 'items', id_)
            return url_join(*uri)

    @staticmethod
    def _get_entity(uri):
        """
        Private function: Parse STA Entity from uri

        :param uri: `str` of STA entity uri

        :returns: `str` of STA Entity
        """
        e = uri.split('/').pop()
        if e in ENTITY:
            return e
        else:
            return ''

    def _generate_mappings(self, provider_def: dict):
        """
        Generate mappings for the STA entity and set up intra-links.

        This function sets up the necessary mappings and configurations for
        the STA entity based on the provided provider definition.

        :param provider_def: `dict` of provider definition containing
                            configuration details for the STA entity.
        """
        self.data.rstrip('/')
        try:
            self.entity = provider_def['entity']
            self._url = url_join(self.data, self.entity)
        except KeyError:
            LOGGER.debug('Attempting to parse Entity from provider data')
            if not self._get_entity(self.data):
                raise RuntimeError('Entity type required')
            self.entity = self._get_entity(self.data)
            self._url = self.data
            self.data = self._url.rstrip(f'/{self.entity}')

        self.parsed_url = urlparse(self.data)

        # Default id
        if self.id_field:
            LOGGER.debug(f'Using id field: {self.id_field}')
        else:
            LOGGER.debug('Using default @iot.id for id field')
            self.id_field = '@iot.id'

        # Custom expand
        if provider_def.get('expand'):
            EXPAND[self.entity] = provider_def['expand']

        # Create intra-links
        self.intralink = provider_def.get('intralink', False)
        if self.intralink and provider_def.get('rel_link'):
            # For pytest
            self.rel_link = provider_def['rel_link']

        elif self.intralink:
            # Read from pygeoapi config
            CONFIG = get_config()
            self.rel_link = get_base_url(CONFIG)

            for name, rs in CONFIG['resources'].items():
                pvs = rs.get('providers')
                p = get_provider_default(pvs)
                e = p.get('entity') or self._get_entity(p['data'])
                if any([
                    not pvs,  # No providers in resource
                    not p.get('intralink'),  # No configuration for intralinks
                    not e,  # No STA entity found
                    self.data not in p.get('data')  # No common STA endpoint
                ]):
                    continue

                if p.get('uri_field'):
                    LOGGER.debug(f'Linking {e} with field: {p["uri_field"]}')
                else:
                    LOGGER.debug(f'Linking {e} with collection: {name}')

                self.linked_entity[e] = {
                    'cnm': name,  # OAPI collection name,
                    'cid': p.get('id_field', '@iot.id'),  # OAPI id_field
                    'uri': p.get('uri_field')  # STA uri_field
                }

    def __repr__(self):
        return f'<SensorThingsProvider> {self.data}, {self.entity}'



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
from functools import reduce
import operator
import os
import re
from typing import Union
import uuid

from dateutil.parser import parse as parse_date
from shapely.geometry import shape
from tinydb import TinyDB, Query, where

from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import (BaseProvider, ProviderConnectionError,
                                    ProviderInvalidQueryError,
                                    ProviderItemNotFoundError)
from pygeoapi.util import get_typed_value

LOGGER = logging.getLogger(__name__)


class TinyDBProvider(BaseProvider):
    """TinyDB Provider"""

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.tinydb_.TinyDBProvider
        """

        super().__init__(provider_def)

        self._excludes = []

        LOGGER.debug(f'Connecting to TinyDB db at {self.data}')

        if not os.path.exists(self.data):
            msg = ('TinyDB file does not exist, please check '
                   'the path specified, or include the full '
                   'path to the file')
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        LOGGER.debug('Checking TinyDB DB permissions')
        if not os.access(self.data, os.W_OK):
            self.db = TinyDB(self.data, access_mode='r')
        else:
            self.db = TinyDB(self.data)

        self.get_fields()

    def get_fields(self):
        """
         Get provider field information (names, types)

        :returns: dict of fields
        """

        if not self._fields:
            try:
                r = self.db.all()[0]
            except IndexError as err:
                LOGGER.debug(err)
                return {}

            for key, value in r['properties'].items():
                if key not in self._excludes:
                    typed_value = get_typed_value(str(value))
                    if isinstance(typed_value, float):
                        typed_value_type = 'number'
                    elif isinstance(typed_value, int):
                        typed_value_type = 'integer'
                    else:
                        typed_value_type = 'string'

                    self._fields[key] = {'type': typed_value_type}

                    try:
                        LOGGER.debug('Attempting to detect date types')
                        _ = parse_date(value)
                        if len(value) > 11:
                            self._fields[key]['format'] = 'date-time'
                        else:
                            self._fields[key]['format'] = 'date'
                    except Exception:
                        LOGGER.debug('No date types detected')
                        pass

        return self._fields

    def get_domains(self, properties=[], current=False) -> tuple:
        records = self.query()
        domains = {}

        if properties:
            keys = properties
        else:
            keys = records['features'][0]['properties'].keys()

        for key in keys:
            v = [x['properties'][key] for x in records['features']]
            v = set([v2 for v2 in v if v2 is not None and isinstance(v2, (float, int, str))])  # noqa

            if v:
                domains[key] = list(v)

        return domains, True

    @crs_transform
    def query(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None, **kwargs):
        """
        query TinyDB document store

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)

        :returns: dict of 0..n GeoJSON feature collection
        """

        Q = Query()
        predicates = []

        feature_collection = {
            'type': 'FeatureCollection',
            'features': []
        }

        if resulttype == 'hits':
            LOGGER.debug('hits only specified')
            limit = 0

        if bbox:
            LOGGER.debug('processing bbox parameter')
            bbox_as_string = ','.join(str(s) for s in bbox)
            predicates.append(Q.geometry.test(bbox_intersects, bbox_as_string))

        if datetime_ is not None:
            LOGGER.debug('processing datetime parameter')
            if self.time_field is None:
                LOGGER.error('time_field not enabled for collection')
                LOGGER.error('Using default time property')
                time_field2 = Q.time
            else:
                LOGGER.error(f'Using properties.{self.time_field}')
                time_field2 = getattr(Q.properties, self.time_field)

            if '/' in datetime_:  # envelope
                LOGGER.debug('detected time range')
                time_begin, time_end = datetime_.split('/')

                if time_begin != '..':
                    predicates.append(time_field2 >= time_begin)
                if time_end != '..':
                    predicates.append(time_field2 <= time_end)

            else:  # time instant
                LOGGER.debug('detected time instant')
                predicates.append(getattr(Q, time_field2) == datetime_)

        if properties:
            LOGGER.debug('processing properties')
            for prop in properties:
                if prop[0] not in self.fields:
                    msg = 'Invalid query: invalid property name'
                    LOGGER.error(msg)
                    raise ProviderInvalidQueryError(msg)

                predicates.append(getattr(Q.properties, prop[0]) == prop[1])

        PQ = reduce(operator.and_, predicates) if predicates else None
        if q:
            SQ = self._add_search_query(Q, q)
        else:
            SQ = None

        try:
            if PQ and SQ:
                results = self.db.search(PQ & SQ)
            elif PQ and not SQ:
                results = self.db.search(PQ)
            elif not PQ and SQ is not None:
                results = self.db.search(SQ)
            else:
                results = self.db.all()
        except SyntaxError as err:
            msg = 'Invalid query'
            LOGGER.error(f'{msg}: {err}')
            raise ProviderInvalidQueryError(msg)

        feature_collection['numberMatched'] = len(results)

        if resulttype == 'hits':
            return feature_collection

        for r in results:
            for e in self._excludes:
                try:
                    del r['properties'][e]
                except KeyError:
                    LOGGER.debug(f'Missing excluded property {e}')

        len_results = len(results)

        LOGGER.debug(f'Results found: {len_results}')

        if len_results > limit:
            returned = limit
        else:
            returned = len_results

        feature_collection['numberReturned'] = returned

        if sortby:
            LOGGER.debug('Sorting results')
            if sortby[0]['order'] == '-':
                sort_reverse = True
            else:
                sort_reverse = False

            results.sort(key=lambda k: k['properties'][sortby[0]['property']],
                         reverse=sort_reverse)

        feature_collection['features'] = results[offset:offset + limit]

        return feature_collection

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        Get TinyDB document by id

        :param identifier: record id

        :returns: `dict` of single record
        """

        LOGGER.debug(f'Fetching identifier {identifier}')

        record = self.db.get(Query().id == identifier)

        if record is None:
            raise ProviderItemNotFoundError('record does not exist')

        for e in self._excludes:
            try:
                del record['properties'][e]
            except KeyError:
                LOGGER.debug(f'Missing excluded property {e}')

        return record

    def create(self, item):
        """
        Adds an item to the TinyDB repository

        :param item: item data

        :returns: identifier of newly created item
        """

        identifier, json_data = self._load_and_prepare_item(
            item, accept_missing_identifier=True)
        if identifier is None:
            # If there is no incoming identifier, allocate a random one
            identifier = str(uuid.uuid4())
            json_data["id"] = identifier

        json_data = self._add_extra_fields(json_data)

        LOGGER.debug(f'Inserting data with identifier {identifier}')
        result = self.db.insert(json_data)

        LOGGER.debug(f'Item added with internal id {result}')

        return identifier

    def update(self, identifier, item):
        """
        Updates an existing item

        :param identifier: feature id
        :param item: `dict` of partial or full item

        :returns: `bool` of update result
        """

        LOGGER.debug(f'Updating item {identifier}')
        identifier, json_data = self._load_and_prepare_item(
            item, identifier, raise_if_exists=False)
        self.db.update(json_data, where('id') == identifier)

        return True

    def delete(self, identifier):
        """
        Deletes an existing item

        :param identifier: item id

        :returns: `bool` of deletion result
        """

        LOGGER.debug(f'Deleting item {identifier}')
        self.db.remove(where('id') == identifier)

        return True

    def _add_extra_fields(self, json_data: dict) -> dict:
        """
        Helper function to add extra fields to an item payload

        :param json_data: `dict` of JSON data

        :returns: `dict` of updated JSON data
        """

        return json_data

    def _add_search_query(self, search_object,
                          search_term: str = None) -> Union[str, None]:
        """
        Create a search query according to the OGC API - Records specification.

        https://docs.ogc.org/is/20-004r1/20-004r1.html (Listing 14)

        Examples (f is shorthand for Q.properties["_metadata-anytext"]):
        +-------------+-----------------------------------+
        | search term | TinyDB search                     |
        +-------------+-----------------------------------+
        | 'aa'        | f.search('aa')                    |
        | 'aa,bb'     | f.search('aa')|f.search('bb')     |
        | 'aa,bb cc'  | f.search('aa')|f.search('bb +cc') |
        +-------------+-----------------------------------+

        :param Q: TinyDB search object
        :param s: `str` of q parameter value

        :returns: `Query` object or `None`
        """

        return search_object

    def __repr__(self):
        return f'<TinyDBProvider> {self.data}'


class TinyDBCatalogueProvider(TinyDBProvider):
    """TinyDB Catalogue Provider"""

    def __init__(self, provider_def):
        super().__init__(provider_def)

        LOGGER.debug('Refreshing fields')
        self._excludes = ['_metadata-anytext']
        self._fields = {}
        self.get_fields()

    def get_fields(self):
        fields = super().get_fields()

        fields['q'] = {'type': 'string'}

        return fields

    def _add_extra_fields(self, json_data: dict) -> dict:
        LOGGER.debug('Adding catalogue anytext property')
        try:
            json_data['properties']['_metadata-anytext'] = ''.join([
                json_data['properties']['title'],
                json_data['properties']['description']
            ])
        except KeyError:
            LOGGER.debug('Missing title and description')
            json_data['properties']['_metadata_anytext'] = ''

        return json_data

    def _prepare_q_param_with_spaces(self, Q: Query, s: str) -> str:
        """
        Prepare a search statement for the search term `s`.
        The term `s` might have spaces.

        Examples (f is shorthand for Q.properties["_metadata-anytext"]):
        +---------------+--------------------+
        | search term   | TinyDB search      |
        +---------------+--------------------+
        | 'aa'          | f.search('aa')     |
        | 'aa bb'       | f.search('aa +bb') |
        | '  aa   bb  ' | f.search('aa +bb') |
        +---------------+--------------------+

        :param Q: TinyDB `Query` object
        :param s: `str` of q parameter value

        :returns: `Query` object
        """

        return Q.properties["_metadata-anytext"].search(
            ' +'.join(s.split()), flags=re.IGNORECASE)

    def _add_search_query(self, search_object,
                          search_term: str = None) -> Union[str, None]:
        """
        Create a search query according to the OGC API - Records specification.

        https://docs.ogc.org/is/20-004r1/20-004r1.html (Listing 14)

        Examples (f is shorthand for Q.properties["_metadata-anytext"]):
        +-------------+-----------------------------------+
        | search term | TinyDB search                     |
        +-------------+-----------------------------------+
        | 'aa'        | f.search('aa')                    |
        | 'aa,bb'     | f.search('aa')|f.search('bb')     |
        | 'aa,bb cc'  | f.search('aa')|f.search('bb +cc') |
        +-------------+-----------------------------------+

        :param Q: TinyDB search object
        :param s: `str` of q parameter value

        :returns: `Query` object or `None`
        """

        if search_term is not None and len(search_term) > 0:
            LOGGER.debug('catalogue q= query')
            terms = [s for s in search_term.split(',') if len(s) > 0]
            terms2 = [self._prepare_q_param_with_spaces(search_object, t)
                      for t in terms]

            return reduce(operator.or_, terms2)
        else:
            return None

    def __repr__(self):
        return f'<TinyDBCatalogueProvider> {self.data}'


def bbox_intersects(record_geometry, input_bbox) -> bool:
    """
    Manual bbox intersection calculation

    :param record_geometry: `dict` of polygon geometry
    :param input_bbox: `str` of 'minx,miny,maxx,maxy'

    :returns: `bool` of whether the record_bbox intersects input_bbox
    """

    if record_geometry is None:
        LOGGER.debug('Record geometry is none; skipping')
        return False

    try:
        bbox1 = list(shape(record_geometry).bounds)
    except Exception as err:
        LOGGER.debug(f'Invalid geometry: {err}')
        return False

    bbox2 = [float(c) for c in input_bbox.split(',')]

    LOGGER.debug(f'Record bbox: {bbox1}')
    LOGGER.debug(f'Input bbox: {bbox2}')

    bbox1_minx, bbox1_miny, bbox1_maxx, bbox1_maxy = bbox1
    bbox2_minx, bbox2_miny, bbox2_maxx, bbox2_maxy = bbox2

    return bbox1_minx <= bbox2_maxx and \
        bbox1_miny <= bbox2_maxy and \
        bbox2_minx <= bbox1_maxx and \
        bbox2_miny <= bbox1_maxy



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Just van den Broecke <justb4@gmail.com>
#
# Copyright (c) 2025 Tom Kralidis
# Copyright (c) 2025 Just van den Broecke
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

"""Generic CRS functions used in the code"""

from copy import deepcopy
import functools
from functools import partial
from dataclasses import dataclass
import logging
from typing import Union, Optional, Callable

import pyproj
import pygeofilter.ast
import pygeofilter.values
from pyproj.exceptions import CRSError
from shapely import ops, Geometry
from shapely.geometry import (
    shape as geojson_to_geom,
    mapping as geom_to_geojson
)

LOGGER = logging.getLogger(__name__)

DEFAULT_CRS_LIST = [
    'http://www.opengis.net/def/crs/OGC/1.3/CRS84',
    'http://www.opengis.net/def/crs/OGC/1.3/CRS84h',
]

DEFAULT_CRS = 'http://www.opengis.net/def/crs/OGC/1.3/CRS84'
DEFAULT_STORAGE_CRS = DEFAULT_CRS


@dataclass
class CrsTransformSpec:
    source_crs_uri: str
    source_crs_wkt: str
    target_crs_uri: str
    target_crs_wkt: str
    always_xy: bool = False


def get_srid(crs: Union[str, pyproj.CRS]) -> Union[int, None]:
    """
    Helper function to attempt to extract an EPSG SRID from
    a `pyproj.CRS` object.

    :param crs: `pyproj.CRS` object

    :returns: int of EPSG SRID, if found
    """
    if isinstance(crs, str):
        crs = get_crs(crs)

    if crs.to_epsg():
        return crs.to_epsg()

    try:
        return pyproj.CRS(crs.to_proj4()).to_epsg()
    except KeyError:
        LOGGER.debug('Unable to extract SRID from proj4 string')


def get_supported_crs_list(
    provider_def: dict, default_crs_list: list = DEFAULT_CRS_LIST
) -> list:
    """
    Helper function to get a complete list of supported CRSs
    from a (Provider) config dict. Result should always include
    a default CRS according to OAPIF Part 2 OGC Standard.
    This will be the default when no CRS list in config or
    added when (partially) missing in config.

    Author: @justb4

    :param provider_def: dictionary with or without a list of CRSs
    :param default_crs_list: default CRS alternatives, first is default

    :returns: list of supported CRSs
    """
    supported_crs_list = provider_def.get('crs', list())
    contains_default = False
    for uri in supported_crs_list:
        if uri in default_crs_list:
            contains_default = True
            break

    # A default CRS is missing: add the first which is the default
    if not contains_default:
        supported_crs_list.append(default_crs_list[0])

    return supported_crs_list


def get_crs(crs: Union[str, pyproj.CRS]) -> pyproj.CRS:
    """
    Get a `pyproj.CRS` instance from a CRS.
    Author: @MTachon

    :param crs: Uniform resource identifier of the coordinate
                reference system. In accordance with
                https://docs.ogc.org/pol/09-048r5.html#_naming_rule
                URIs can take either the form of a URL or a URN
                or `pyproj.CRS` object
    :raises `CRSError`: Error raised if no CRS could be identified from the
        URI.

    :returns: `pyproj.CRS` instance matching the input URI.
    """

    if isinstance(crs, pyproj.CRS):
        return crs

    # normalize the input `uri` to a URL first
    uri = str(crs)
    url = uri.replace(
        'urn:ogc:def:crs', 'http://www.opengis.net/def/crs'
    ).replace(':', '/')
    try:
        authority, code = url.rsplit('/', maxsplit=3)[1::2]
        crs = pyproj.CRS.from_authority(authority, code)
    except ValueError:
        msg = (
            f'CRS could not be identified from URI {uri!r}. CRS URIs must '
            'follow one of two formats: '
            '"http://www.opengis.net/def/crs/{authority}/{version}/{code}" or '
            '"urn:ogc:def:crs:{authority}:{version}:{code}" '
            '(see https://docs.opengeospatial.org/is/18-058r1/18-058r1.html#crs-overview).'  # noqa
        )
        LOGGER.error(msg)
        raise CRSError(msg)
    except CRSError:
        msg = f"CRS could not be identified from URI {uri!r}"
        LOGGER.error(msg)
        raise CRSError(msg)

    return crs


def get_transform_from_spec(
    crs_transform_spec: CrsTransformSpec
) -> Callable[[Geometry], Geometry]:
    """ Get transformation function from a `CrsTransformSpec` instance.

    :param crs_transform_spec: `CrsTransformSpec`

    :returns: `callable` Function to transform the coordinates of a `Geometry`.
    """
    if crs_transform_spec is None:
        LOGGER.warning('No transform spec found')
        return None

    return get_transform_from_crs(
        pyproj.CRS.from_wkt(crs_transform_spec.source_crs_wkt),
        pyproj.CRS.from_wkt(crs_transform_spec.target_crs_wkt),
        crs_transform_spec.always_xy
    )


def get_transform_from_crs(
    crs_in: pyproj.CRS, crs_out: pyproj.CRS, always_xy: bool = False
) -> Callable[[Geometry], Geometry]:
    """ Get transformation function from two `pyproj.CRS` instances.

    Get function to transform the coordinates of a Shapely geometrical object
    from one coordinate reference system to another.

    :param crs_in: `pyproj.CRS` Input Coordinate Reference System
    :param crs_out: `pyproj.CRS` Output Coordinate Reference System
    :param always_xy: 'bool' should axis order be forced to x,y (lon, lat)
                       even if CRSdeclares y,x (lat,lon)

    :returns: `callable` Function to transform the coordinates of a `Geometry`.
    """
    crs_transform = pyproj.Transformer.from_crs(
        crs_in, crs_out, always_xy=always_xy,
    ).transform
    return partial(ops.transform, crs_transform)


def crs_transform(func):
    """Decorator that transforms the geometry's/geometries' coordinates of a
    Feature/FeatureCollection.

    This function can be used to decorate another function which returns either
    a Feature or a FeatureCollection (GeoJSON-like `dict`). For a
    FeatureCollection, the Features are stored in a ´list´ available at the
    'features' key of the returned `dict`. For each Feature, the geometry is
    available at the 'geometry' key. The decorated function may take a
    'crs_transform_spec' parameter, which accepts a `CrsTransformSpec` instance
    as value. If the `CrsTransformSpec` instance represents a coordinates
    transformation between two different CRSs, the coordinates of the
    Feature's/FeatureCollection's geometry/geometries will be transformed
    before returning the Feature/FeatureCollection. If the 'crs_transform_spec'
    parameter is not given, passed `None` or passed a `CrsTransformSpec`
    instance which does not represent a coordinates transformation, the
    Feature/FeatureCollection is returned unchanged. This decorator can for
    example be use to help supporting coordinates transformation of
    Feature/FeatureCollection `dict` objects returned by the `get` and `query`
    methods of (new or with no native support for transformations) providers of
    type 'feature'.

    :param func: Function to decorate.

    :returns: Decorated function.
    """
    @functools.wraps(func)
    def get_geojsonf(*args, **kwargs):
        crs_transform_spec = kwargs.get('crs_transform_spec')
        result = func(*args, **kwargs)
        if crs_transform_spec is None:
            # No coordinates transformation for feature(s) returned by the
            # decorated function.
            LOGGER.debug('crs_transform: NOT applying coordinate transforms')
            return result
        # Create transformation function and transform the output feature(s)'
        # coordinates before returning them.
        transform_func = get_transform_from_spec(crs_transform_spec)
        LOGGER.debug(f'crs_transform: transforming features CRS '
                     f'from {crs_transform_spec.source_crs_uri} '
                     f'to {crs_transform_spec.target_crs_uri}')

        features = result.get('features')
        # Decorated function returns a single Feature
        if features is None:
            # Transform the feature's coordinates
            crs_transform_feature(result, transform_func)
        # Decorated function returns a FeatureCollection
        else:
            # Transform all features' coordinates
            for feature in features:
                crs_transform_feature(feature, transform_func)
        return result
    return get_geojsonf


def crs_transform_feature(feature: dict, transform_func: Callable):
    """Transform the coordinates of a Feature.

    :param feature: Feature (GeoJSON-like `dict`) to transform.
    :param transform_func: Function that transforms the coordinates of a
        `Geometry` instance.

    :returns: None
    """
    json_geometry = feature.get('geometry')
    if json_geometry is not None:
        feature['geometry'] = geom_to_geojson(
            transform_func(geojson_to_geom(json_geometry))
        )


def transform_bbox(bbox: list, from_crs: Union[str, pyproj.CRS],
                   to_crs: Union[str, pyproj.CRS]) -> list:
    """
    helper function to transform a bounding box (bbox) from
    a source to a target CRS. CRSs in URI str format.
    Uses pyproj Transformer.

    :param bbox: list of coordinates in 'from_crs' projection
    :param from_crs: CRS to transform from
    :param to_crs: CRS to transform to
    :raises `CRSError`: Error raised if no CRS could be identified from an
        URI.

    :returns: list of 4 or 6 coordinates
    """

    from_crs_obj = get_crs(from_crs)
    to_crs_obj = get_crs(to_crs)
    transform_func = pyproj.Transformer.from_crs(
        from_crs_obj, to_crs_obj).transform

    n_dims = len(bbox) // 2
    return list(transform_func(*bbox[:n_dims]) + transform_func(
        *bbox[n_dims:]))


def modify_pygeofilter(
    ast_tree: pygeofilter.ast.Node,
    *,
    filter_crs_uri: str,
    storage_crs_uri: Optional[str] = None,
    geometry_column_name: Optional[str] = None
) -> pygeofilter.ast.Node:
    """
    Modifies the input pygeofilter with information from the provider.

    :param ast_tree: `pygeofilter.ast.Node` representing the
                     already parsed pygeofilter expression
    :param filter_crs_uri: URI of the CRS being used in the filtering
                           expression
    :param storage_crs_uri: An optional string containing the URI of
                            the provider's storage CRS
    :param geometry_column_name: An optional string containing the
                                 actual name of the provider's geometry field
    :returns: A new pygeofilter.ast.Node, with the modified filter
              expression

    This function modifies the parsed pygeofilter that contains the raw
    filter expression provided by an external client. It performs the
    following modifications:

    - if the filter includes any spatial coordinates and they are being
      provided in a different CRS from the provider's storage CRS, the
      corresponding geometries are transformed into the storage CRS

    - if the filter includes the generic 'geometry' name as a reference to
      the actual geometry of features, it is replaced by the actual name
      of the geometry field, as specified by the provider

    """
    new_tree = deepcopy(ast_tree)
    if storage_crs_uri:
        _inplace_transform_filter_geometries(
            new_tree, get_crs(filter_crs_uri), get_crs(storage_crs_uri)
        )
    if geometry_column_name:
        _inplace_replace_geometry_filter_name(new_tree, geometry_column_name)
    return new_tree


def _inplace_transform_filter_geometries(
    node: pygeofilter.ast.Node,
    filter_crs: pyproj.CRS,
    storage_crs: pyproj.CRS
) -> None:
    """
    Recursively traverse node tree and convert coordinates to the storage CRS.

    This function modifies nodes in the already-parsed filter in order to find
    any geometry literals that may be used in the filter and, if necessary,
    proceeds to convert spatial coordinates to the CRS used by the provider.
    """
    try:
        sub_nodes = node.get_sub_nodes()
    except AttributeError:
        pass
    else:
        for sub_node in sub_nodes:
            is_geometry_node = isinstance(
                sub_node, pygeofilter.values.Geometry)
            if is_geometry_node:
                # NOTE1: To be flexible, and since pygeofilter
                # already supports it, in addition to supporting
                # the `filter-crs` parameter, we also support having a
                # geometry defined in EWKT, meaning the CRS is provided
                # inline, like this `SRID=<CRS_CODE>;<WKT>` - If provided,
                # this overrides the value of `filter-crs`. This enables
                # supporting, for example, an exotic filter expression with
                # multiple geometries specified in different CRSs

                # NOTE2: We specify a default CRS using a URI of type URN
                # because this is what pygeofilter uses internally too

                crs_urn_provided_in_ewkt = sub_node.geometry.get(
                    'crs', {}).get('properties', {}).get('name')
                if crs_urn_provided_in_ewkt is not None:
                    crs = get_crs(crs_urn_provided_in_ewkt)
                else:
                    crs = filter_crs
                if crs != storage_crs:
                    # convert geometry coordinates to storage crs
                    geom = geojson_to_geom(sub_node.geometry)
                    coord_transformer = pyproj.Transformer.from_crs(
                        crs_from=crs, crs_to=storage_crs).transform
                    transformed_geom = ops.transform(coord_transformer, geom)
                    sub_node.geometry = geom_to_geojson(transformed_geom)
                # ensure the crs is encoded in the sub-node, otherwise
                # pygeofilter will assign it its own default CRS
                authority, code = storage_crs.to_authority()
                sub_node.geometry['crs'] = {
                    'properties': {
                        'name': f'urn:ogc:def:crs:{authority}::{code}'
                    }
                }
            else:
                _inplace_transform_filter_geometries(
                    sub_node, filter_crs, storage_crs)


def _inplace_replace_geometry_filter_name(
    node: pygeofilter.ast.Node,
    geometry_column_name: str
) -> None:
    """Recursively traverse node tree and rename nodes of type ``Attribute``.

    Nodes of type ``Attribute`` named ``geometry`` are renamed to the value of
    the ``geometry_column_name`` parameter.
    """
    try:
        sub_nodes = node.get_sub_nodes()
    except AttributeError:
        pass
    else:
        for sub_node in sub_nodes:
            is_attribute_node = isinstance(sub_node, pygeofilter.ast.Attribute)
            if is_attribute_node and sub_node.name == "geometry":
                sub_node.name = geometry_column_name
            else:
                _inplace_replace_geometry_filter_name(
                    sub_node, geometry_column_name)


def create_crs_transform_spec(
    provider_def: dict, query_crs_uri: Optional[str] = None
) -> Union[None, CrsTransformSpec]:
    """
    Create a `CrsTransformSpec` instance based on provider config and
    *crs* query parameter.

    :param provider_def: Provider config dictionary.
    :param query_crs_uri: Uniform resource identifier of the coordinate
        reference system (CRS) specified in query parameter (if specified).

    :raises ValueError: Error raised if the CRS specified in the query
        parameter is not in the list of supported CRSs of the provider.
    :raises `CRSError`: Error raised if no CRS could be identified from the
        query *crs* parameter (URI).

    :returns: `CrsTransformSpec` instance if the CRS specified in query
        parameter differs from the storage CRS, else `None`.
    """

    # Get storage/default CRS for Collection.
    always_xy = provider_def.get('always_xy', False)
    storage_crs_uri = provider_def.get('storage_crs', DEFAULT_STORAGE_CRS)
    storage_crs = get_crs(storage_crs_uri)

    if not query_crs_uri:
        if storage_crs_uri in DEFAULT_CRS_LIST:
            # Could be that storage_crs is
            # http://www.opengis.net/def/crs/OGC/1.3/CRS84h
            query_crs_uri = storage_crs_uri
        else:
            query_crs_uri = DEFAULT_CRS
        LOGGER.debug(f'no crs parameter, using default: {query_crs_uri}')

    supported_crs_list = get_supported_crs_list(provider_def)
    # Check that the crs specified by the query parameter is supported.
    if query_crs_uri not in supported_crs_list:
        raise ValueError(
            f'CRS {query_crs_uri!r} not supported for this '
            'collection. List of supported CRSs: '
            f'{", ".join(supported_crs_list)}.'
        )
    crs_out = get_crs(query_crs_uri)

    # Check if the crs specified in query parameter differs from the
    # storage crs.
    if storage_crs == crs_out:
        LOGGER.debug('No CRS transformation')
        return None

    LOGGER.debug(
        f'CRS transformation: {storage_crs} -> {crs_out}'
    )
    return CrsTransformSpec(
        source_crs_uri=storage_crs_uri,
        source_crs_wkt=storage_crs.to_wkt(),
        target_crs_uri=query_crs_uri,
        target_crs_wkt=crs_out.to_wkt(),
        always_xy=always_xy
    )


def set_content_crs_header(
    headers: dict, config: dict, query_crs_uri: Optional[str] = None,
) -> None:
    """Set the *Content-Crs* header in responses from providers of Feature
    type.

    :param headers: Response headers dictionary.
    :param config: Provider config dictionary.
    :param query_crs_uri: Uniform resource identifier of the coordinate
        reference system specified in query parameter (if specified).

    :returns: None
    """

    if query_crs_uri:
        content_crs_uri = query_crs_uri
    else:
        # If empty use default CRS
        storage_crs_uri = config.get('storage_crs', DEFAULT_STORAGE_CRS)
        if storage_crs_uri in DEFAULT_CRS_LIST:
            content_crs_uri = storage_crs_uri
        else:
            content_crs_uri = DEFAULT_CRS

    headers['Content-Crs'] = f'<{content_crs_uri}>'



# =================================================================
#
# Authors: Jorge Samuel Mendes de Jesus <jorge.dejesus@protonmail.com>
#          Tom Kralidis <tomkralidis@gmail.com>
#          Mary Bucknell <mbucknell@usgs.gov>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#          Colton Loftus <cloftus@lincolninst.edu>
#
# Copyright (c) 2018 Jorge Samuel Mendes de Jesus
# Copyright (c) 2025 Tom Kralidis
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2024 Bernhard Mallinger
# Copyright (c) 2025 Colton Loftus
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
from datetime import datetime
from decimal import Decimal
import functools
import logging
from typing import Optional, Any

from geoalchemy2 import Geometry  # noqa - this isn't used explicitly but is needed to process Geometry columns
from geoalchemy2.functions import ST_MakeEnvelope, ST_Intersects
from geoalchemy2.shape import to_shape, from_shape
from pygeofilter.backends.sqlalchemy.evaluate import to_filter
import shapely
from sqlalchemy.sql import func
from sqlalchemy import (
    create_engine,
    MetaData,
    PrimaryKeyConstraint,
    asc,
    desc,
    delete
)
from sqlalchemy.engine import URL, Engine
from sqlalchemy.exc import (
    ConstraintColumnNotFoundError,
    InvalidRequestError,
    OperationalError,
    SQLAlchemyError
)
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session, load_only
from sqlalchemy.sql.expression import and_
from sqlalchemy.schema import Table

from pygeoapi.crs import get_transform_from_spec, get_srid
from pygeoapi.provider.base import (
    BaseProvider,
    ProviderConnectionError,
    ProviderInvalidDataError,
    ProviderQueryError,
    ProviderItemNotFoundError
)

LOGGER = logging.getLogger(__name__)


class GenericSQLProvider(BaseProvider):
    """
    Generic provider for sql databases it can be inherited
    from to create specific providers for different databases
    """

    def __init__(
        self,
        provider_def: dict,
        driver_name: str,
        extra_conn_args: Optional[dict] = {}
    ):
        """
        GenericSQLProvider Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class
                             data contains the connection information
                             for class DatabaseCursor
        :param driver_name: database driver name
        :param extra_conn_args: additional custom connection arguments to
                                pass for a query

        :returns: pygeoapi.provider.GenericSQLProvider
        """
        LOGGER.debug('Initialising GenericSQL provider.')
        super().__init__(provider_def)

        self.table = provider_def['table']
        self.id_field = provider_def['id_field']
        self.geom = provider_def.get('geom_field', 'geom')
        self.driver_name = driver_name

        LOGGER.debug(f'Name: {self.name}')
        LOGGER.debug(f'Table: {self.table}')
        LOGGER.debug(f'ID field: {self.id_field}')
        LOGGER.debug(f'Geometry field: {self.geom}')
        LOGGER.debug(f'Configured Storage CRS: {self.storage_crs}')

        # Read table information from database
        options = provider_def.get('options', {}) | extra_conn_args
        store_db_parameters(self, provider_def['data'], options)
        self._engine = get_engine(
            driver_name,
            self.db_host,
            self.db_port,
            self.db_name,
            self.db_user,
            self._db_password,
            self.db_conn,
            **self.db_options
        )
        self.table_model = get_table_model(
            self.table, self.id_field, self.db_search_path, self._engine
        )

        self.get_fields()

    def query(
        self,
        offset=0,
        limit=10,
        resulttype='results',
        bbox=[],
        datetime_=None,
        properties=[],
        sortby=[],
        select_properties=[],
        skip_geometry=False,
        q=None,
        filterq=None,
        crs_transform_spec=None,
        **kwargs
    ):
        """
        Query sql database for all the content.
        e,g: http://localhost:5000/collections/hotosm_bdi_waterways/items?
        limit=1&resulttype=results

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)
        :param filterq: CQL query as text string
        :param crs_transform_spec: `CrsTransformSpec` instance, optional

        :returns: GeoJSON FeatureCollection
        """

        LOGGER.debug('Preparing filters')
        property_filters = self._get_property_filters(properties)
        cql_filters = self._get_cql_filters(filterq)
        bbox_filter = self._get_bbox_filter(bbox)
        time_filter = self._get_datetime_filter(datetime_)
        order_by_clauses = self._get_order_by_clauses(sortby, self.table_model)
        selected_properties = self._select_properties_clause(
            select_properties, skip_geometry
        )

        LOGGER.debug('Querying Database')
        # Execute query within self-closing database Session context
        with Session(self._engine) as session:
            results = (
                session.query(self.table_model)
                .filter(property_filters)
                .filter(cql_filters)
                .filter(bbox_filter)
                .filter(time_filter)
                .options(selected_properties)
            )

            LOGGER.debug('Preparing response')
            response = {
                'type': 'FeatureCollection',
                'features': [],
                'numberReturned': 0
            }

            if self.count or resulttype == 'hits':
                matched = results.count()
                response['numberMatched'] = matched
                LOGGER.debug(f'Found {matched} result(s)')
            else:
                LOGGER.debug('Count disabled')

            if resulttype == 'hits' or not results:
                return response

            crs_transform_out = get_transform_from_spec(crs_transform_spec)

            for item in (
                results.order_by(*order_by_clauses).offset(offset).limit(limit)
            ):
                response['numberReturned'] += 1
                response['features'].append(
                    self._sqlalchemy_to_feature(item, crs_transform_out,
                                                select_properties)
                )

        return response

    def get_fields(self):
        """
        Return fields (columns) from database table

        :returns: dict of fields
        """

        LOGGER.debug('Get available fields/properties')

        # sql-schema only allows these types, so we need to map from sqlalchemy
        # string, number, integer, object, array, boolean, null,
        # https://json-schema.org/understanding-json-schema/reference/type.html
        column_type_map = {
            bool: 'boolean',
            datetime: 'string',
            Decimal: 'number',
            dict: 'object',
            float: 'number',
            int: 'integer',
            str: 'string'
        }
        default_type = 'string'

        # https://json-schema.org/understanding-json-schema/reference/string#built-in-formats  # noqa
        column_format_map = {
            'date': 'date',
            'interval': 'duration',
            'time': 'time',
            'timestamp': 'date-time'
        }

        def _column_type_to_json_schema_type(column_type):
            try:
                python_type = column_type.python_type
            except NotImplementedError:
                LOGGER.warning(f'Unsupported column type {column_type}')
                return default_type
            else:
                try:
                    return column_type_map[python_type]
                except KeyError:
                    LOGGER.warning(f'Unsupported column type {column_type}')
                    return default_type

        def _column_format_to_json_schema_format(column_type):
            try:
                ct = str(column_type).lower()
                return column_format_map[ct]
            except KeyError:
                LOGGER.debug('No string format detected')
                return None

        if not self._fields:
            for column in self.table_model.__table__.columns:
                LOGGER.debug(f'Testing {column.name}')
                if column.name == self.geom:
                    continue

                self._fields[str(column.name)] = {
                    'type': _column_type_to_json_schema_type(column.type),
                    'format': _column_format_to_json_schema_format(
                        column.type
                    )
                }

        return self._fields

    def get(self, identifier, crs_transform_spec=None, **kwargs):
        """
        Query the provider for a specific
        feature id e.g: /collections/hotosm_bdi_waterways/items/13990765

        :param identifier: feature id
        :param crs_transform_spec: `CrsTransformSpec` instance, optional

        :returns: GeoJSON FeatureCollection
        """
        LOGGER.debug(f'Get item by ID: {identifier}')

        # Execute query within self-closing database Session context
        with Session(self._engine) as session:
            # Retrieve data from database as feature
            try:
                item = session.get(self.table_model, identifier)
                # Ensure that item is not None
                assert item is not None
            except (AssertionError, SQLAlchemyError) as e:
                LOGGER.debug(e, exc_info=True)
                msg = f'No such item: {self.id_field}={identifier}.'
                raise ProviderItemNotFoundError(msg)
            crs_transform_out = get_transform_from_spec(crs_transform_spec)
            feature = self._sqlalchemy_to_feature(item, crs_transform_out)

            # Drop non-defined properties
            if self.properties:
                props = feature['properties']
                dropping_keys = deepcopy(props).keys()
                for item in dropping_keys:
                    if item not in self.properties:
                        props.pop(item)

            # Add fields for previous and next items
            id_field = getattr(self.table_model, self.id_field)
            prev_item = (
                session.query(self.table_model)
                .order_by(id_field.desc())
                .filter(id_field < identifier)
                .first()
            )
            next_item = (
                session.query(self.table_model)
                .order_by(id_field.asc())
                .filter(id_field > identifier)
                .first()
            )
            feature['prev'] = (
                getattr(prev_item, self.id_field)
                if prev_item is not None
                else identifier
            )
            feature['next'] = (
                getattr(next_item, self.id_field)
                if next_item is not None
                else identifier
            )

        return feature

    def create(self, item):
        """
        Create a new item

        :param item: `dict` of new item

        :returns: identifier of created item
        """

        identifier, json_data = self._load_and_prepare_item(
            item, accept_missing_identifier=True
        )

        new_instance = self._feature_to_sqlalchemy(json_data, identifier)
        with Session(self._engine) as session:
            session.add(new_instance)
            session.commit()
            result_id = getattr(new_instance, self.id_field)

        # NOTE: need to use id from instance in case it's generated
        return result_id

    def update(self, identifier, item):
        """
        Updates an existing item

        :param identifier: feature id
        :param item: `dict` of partial or full item

        :returns: `bool` of update result
        """

        identifier, json_data = self._load_and_prepare_item(
            item, raise_if_exists=False
        )

        new_instance = self._feature_to_sqlalchemy(json_data, identifier)
        with Session(self._engine) as session:
            session.merge(new_instance)
            session.commit()

        return True

    def delete(self, identifier):
        """
        Deletes an existing item

        :param identifier: item id

        :returns: `bool` of deletion result
        """
        with Session(self._engine) as session:
            id_column = getattr(self.table_model, self.id_field)
            result = session.execute(
                delete(self.table_model).where(id_column == identifier)
            )
            session.commit()

        return result.rowcount > 0

    def _sqlalchemy_to_feature(self, item, crs_transform_out=None,
                               select_properties=[]):
        """
        Helper function to transform an SQLAlchemy result to a
        GeoJSON feature.

        :param item: SQLAlchemy result
        :param crs_transform_out: CRS transformation
        :param select_properties: additional properties to filter on

        :returns: `dict` of GeoJSON feature
        """

        feature = {'type': 'Feature', 'properties': {}}

        item_dict = item.__dict__

        # set feature id
        feature['id'] = item_dict[self.id_field]

        # Convert geometry to GeoJSON style
        if item_dict.get(self.geom) is not None:
            wkb_geom = item_dict[self.geom]
            try:
                shapely_geom = to_shape(wkb_geom)
            except TypeError:
                shapely_geom = shapely.geometry.shape(wkb_geom)
            if crs_transform_out is not None:
                shapely_geom = crs_transform_out(shapely_geom)
            geojson_geom = shapely.geometry.mapping(shapely_geom)
            feature['geometry'] = geojson_geom
        else:
            feature['geometry'] = None

        keys = select_properties or self.fields.keys()
        for key in keys:
            if key in item_dict:
                feature['properties'][key] = item_dict[key]

        return feature

    def _feature_to_sqlalchemy(self, json_data, identifier=None):
        attributes = {**json_data['properties']}
        # 'identifier' key maybe be present in geojson properties, but might
        # not be a valid db field
        attributes.pop('identifier', None)
        attributes[self.geom] = from_shape(
            shapely.geometry.shape(json_data['geometry']),
            srid=get_srid(self.storage_crs)
        )
        attributes[self.id_field] = identifier

        try:
            return self.table_model(**attributes)
        except Exception as e:
            LOGGER.exception('Failed to create db model')
            raise ProviderInvalidDataError(str(e))

    def _get_order_by_clauses(self, sort_by, table_model):
        # Build sort_by clauses if provided
        clauses = []
        for sort_by_dict in sort_by:
            model_column = getattr(table_model, sort_by_dict['property'])
            order_function = asc if sort_by_dict['order'] == '+' else desc
            clauses.append(order_function(model_column))

        # Otherwise sort by primary key (to ensure reproducible output)
        if not clauses:
            clauses.append(asc(getattr(table_model, self.id_field)))

        return clauses

    def _get_cql_filters(self, filterq):
        if not filterq:
            return True  # Let everything through

        # Convert filterq into SQL Alchemy filters
        field_mapping = {
            column_name: getattr(self.table_model, column_name)
            for column_name in self.table_model.__table__.columns.keys()
        }
        cql_filters = to_filter(filterq, field_mapping)

        return cql_filters

    def _get_property_filters(self, properties):
        if not properties:
            return True  # Let everything through

        # Convert property filters into SQL Alchemy filters
        # Based on https://stackoverflow.com/a/14887813/3508733
        filter_group = []
        for column_name, value in properties:
            column = getattr(self.table_model, column_name)
            filter_group.append(column == value)
        property_filters = and_(*filter_group)

        return property_filters

    def _get_bbox_filter(self, bbox: list[float]):
        """
        Construct the bounding box filter function that
        will be used in the query; this is dependent on the
        underlying db driver
        """
        raise NotImplementedError

    def _get_datetime_filter(self, datetime_):
        if datetime_ in (None, '../..'):
            return True
        else:
            if self.time_field is None:
                LOGGER.error('time_field not enabled for collection')
                raise ProviderQueryError()

            time_column = getattr(self.table_model, self.time_field)

            if '/' in datetime_:  # envelope
                LOGGER.debug('detected time range')
                time_begin, time_end = datetime_.split('/')
                if time_begin == '..':
                    datetime_filter = time_column <= time_end
                elif time_end == '..':
                    datetime_filter = time_column >= time_begin
                else:
                    datetime_filter = time_column.between(time_begin, time_end)
            else:
                datetime_filter = time_column == datetime_
        return datetime_filter

    def _select_properties_clause(self, select_properties, skip_geometry):
        # List the column names that we want
        if select_properties:
            column_names = sorted(set(select_properties),
                                  key=select_properties.index)
        else:
            # get_fields() doesn't include geometry column
            column_names = self.fields.keys()

        if self.properties:  # optional subset of properties defined in config
            properties_from_config = self.properties
            column_names = column_names and properties_from_config

        if not skip_geometry:
            column_names = list(column_names)
            column_names.append(self.geom)

        # Convert names to SQL Alchemy clause
        selected_columns = []
        for column_name in column_names:
            try:
                column = getattr(self.table_model, column_name)
                selected_columns.append(column)
            except AttributeError:
                pass  # Ignore non-existent columns
        selected_properties_clause = load_only(*selected_columns)

        return selected_properties_clause


def store_db_parameters(
    self: GenericSQLProvider | Any,
    connection_data: str | dict[str],
    options: dict[str, str]
) -> None:
    """
    Store database connection parameters

    :self: instance of provider or manager class
    :param connection_data: connection string or dict of connection params
    :param options: additional connection options

    :returns: None
    """
    if isinstance(connection_data, str):
        self.db_conn = connection_data
        connection_data = {}
    else:
        self.db_conn = None
    # OR
    self.db_user = connection_data.get('user')
    self.db_host = connection_data.get('host')
    self.db_port = connection_data.get('port', self.default_port)
    self.db_name = (
        connection_data.get('dbname') or connection_data.get('database')
    )
    self.db_query = connection_data.get('query')
    self._db_password = connection_data.get('password')
    # db_search_path gets converted to a tuple here in order to ensure it
    # is hashable - which allows us to use functools.cache() when
    # reflecting the table definition from the DB
    self.db_search_path = tuple(
        connection_data.get('search_path') or
        options.pop('search_path', ['public'])
    )
    self.db_options = {
        k: v
        for k, v in options.items()
        if not isinstance(v, dict)
    }


@functools.cache
def get_engine(
    driver_name: str,
    host: str,
    port: str,
    database: str,
    user: str,
    password: str,
    conn_str: Optional[str] = None,
    **connect_args
) -> Engine:
    """
    Get SQL Alchemy engine.

    :param driver_name: database driver name
    :param host: database host
    :param port: database port
    :param database: database name
    :param user: database user
    :param password: database password
    :param conn_str: optional connection URL
    :param connect_args: custom connection arguments to pass to create_engine()

    :returns: SQL Alchemy engine
    """
    if conn_str is None:
        conn_str = URL.create(
            drivername=driver_name,
            username=user,
            password=password,
            host=host,
            port=int(port),
            database=database
        )

    engine = create_engine(
        conn_str, connect_args=connect_args, pool_pre_ping=True
    )

    LOGGER.debug(f'Created engine for {repr(engine.url)}.')
    return engine


@functools.cache
def get_table_model(
    table_name: str,
    id_field: str,
    db_search_path: tuple[str],
    engine: Engine
) -> Table:
    """
    Reflect table using SQLAlchemy Automap.

    :param table_name: name of table to reflect
    :param id_field: name of primary key field
    :param db_search_path: tuple of database schemas to search for the table
    :param engine: SQLAlchemy engine to use for reflection

    :returns: SQLAlchemy model of the reflected table
    """
    LOGGER.debug('Reflecting table definition from database')
    metadata = MetaData()

    # Look for table in the first schema in the search path
    schema = db_search_path[0]
    try:
        LOGGER.debug(f'Looking for table {table_name} in schema {schema}')
        metadata.reflect(
            bind=engine, schema=schema, only=[table_name], views=True
        )
    except OperationalError:
        raise ProviderConnectionError(
            f'Could not connect to {repr(engine.url)} (password hidden).'
        )
    except InvalidRequestError:
        msg = (
            f"Table '{table_name}' not found in schema '{schema}' "
            f'on {repr(engine.url)}.'
        )
        LOGGER.error(msg)

        if len(db_search_path) > 1:
            # If the table is not found in the first schema, try the next one
            return get_table_model(
                table_name,
                id_field,
                db_search_path[1:],
                engine
            )
        else:
            # If the table is not found in any schema, raise an error
            raise ProviderQueryError(msg)

    # Create SQLAlchemy model from reflected table
    # It is necessary to add the primary key constraint because SQLAlchemy
    # requires it to reflect the table, but a view in a PostgreSQL database
    # does not have a primary key defined.
    sqlalchemy_table_def = metadata.tables[f'{schema}.{table_name}']
    try:
        sqlalchemy_table_def.append_constraint(PrimaryKeyConstraint(id_field))
    except (ConstraintColumnNotFoundError, KeyError):
        raise ProviderQueryError(
            f'No such id_field column ({id_field}) on {schema}.{table_name}.'
        )

    _Base = automap_base(metadata=metadata)
    _Base.prepare(
        name_for_scalar_relationship=_name_for_scalar_relationship,
    )
    return getattr(_Base.classes, table_name)


def _name_for_scalar_relationship(base, local_cls, referred_cls, constraint):
    """Function used when automapping classes and relationships from
    database schema and fixes potential naming conflicts.
    """
    name = referred_cls.__name__.lower()
    local_table = local_cls.__table__
    if name in local_table.columns:
        newname = name + '_'
        LOGGER.debug(
            f'Already detected column name {name!r} in table '
            f'{local_table!r}. Using {newname!r} for relationship name.'
        )
        return newname
    return name


class PostgreSQLProvider(GenericSQLProvider):
    """
    A provider for querying a PostgreSQL database
    """
    default_port = 5432

    def __init__(self, provider_def: dict):
        """
        PostgreSQLProvider Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class
                             data contains the connection information
                             for class DatabaseCursor
        :returns: pygeoapi.provider.sql.PostgreSQLProvider
        """

        driver_name = 'postgresql+psycopg2'
        extra_conn_args = {
            'client_encoding': 'utf8',
            'application_name': 'pygeoapi'
        }
        super().__init__(provider_def, driver_name, extra_conn_args)

    def _get_bbox_filter(self, bbox: list[float]):
        """
        Construct the bounding box filter function
        """
        if not bbox:
            return True  # Let everything through if no bbox

        storage_srid = get_srid(self.storage_crs)
        envelope = ST_MakeEnvelope(*bbox, storage_srid)

        geom_column = getattr(self.table_model, self.geom)
        bbox_filter = ST_Intersects(envelope, geom_column)

        return bbox_filter


class MySQLProvider(GenericSQLProvider):
    """
    A provider for a MySQL database
    """
    default_port = 3306

    def __init__(self, provider_def: dict):
        """
        MySQLProvider Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class
                             data contains the connection information
                             for class DatabaseCursor
        :returns: pygeoapi.provider.sql.MySQLProvider
        """

        driver_name = 'mysql+pymysql'
        extra_conn_args = {
            'charset': 'utf8mb4'
        }
        super().__init__(provider_def, driver_name, extra_conn_args)

    def _get_bbox_filter(self, bbox: list[float]):
        """
        Construct the bounding box filter function
        """
        if not bbox:
            return True  # Let everything through if no bbox

        # If we are using mysql we can't use ST_MakeEnvelope since it is
        # postgis specific and thus we have to use MBRContains with a WKT
        # POLYGON

        # Create WKT POLYGON from bbox: (minx, miny, maxx, maxy)
        minx, miny, maxx, maxy = bbox
        polygon_wkt = f'POLYGON(({minx} {miny}, {maxx} {miny}, {maxx} {maxy}, {minx} {maxy}, {minx} {miny}))'  # noqa
        geom_column = getattr(self.table_model, self.geom)
        # Use MySQL MBRContains for index-accelerated bounding box checks
        bbox_filter = func.MBRContains(
            func.ST_GeomFromText(polygon_wkt), geom_column
        )
        return bbox_filter

    def get(self, identifier, crs_transform_spec=None, **kwargs):
        """
        Query the provider for a specific
        feature id e.g: /collections/hotosm_bdi_waterways/items/13990765

        :param identifier: feature id
        :param crs_transform_spec: `CrsTransformSpec` instance, optional

        :returns: GeoJSON FeatureCollection
        """
        LOGGER.debug(f'Get item by ID: {identifier}')

        # Execute query within self-closing database Session context
        with Session(self._engine) as session:
            # Retrieve data from database as feature
            try:
                item = session.get(self.table_model, identifier)
                # Ensure that item is not None
                assert item is not None
                # Ensure returned row has exact match
                feature_id = getattr(item, self.id_field)
                assert str(feature_id) == identifier
            except (AssertionError, SQLAlchemyError) as e:
                LOGGER.debug(e, exc_info=True)
                msg = f'No such item: {self.id_field}={identifier}.'
                raise ProviderItemNotFoundError(msg)
            crs_transform_out = get_transform_from_spec(crs_transform_spec)
            feature = self._sqlalchemy_to_feature(item, crs_transform_out)

            # Drop non-defined properties
            if self.properties:
                props = feature['properties']
                dropping_keys = deepcopy(props).keys()
                for item in dropping_keys:
                    if item not in self.properties:
                        props.pop(item)

            # Add fields for previous and next items
            id_field = getattr(self.table_model, self.id_field)
            prev_item = (
                session.query(self.table_model)
                .order_by(id_field.desc())
                .filter(id_field < feature_id)
                .first()
            )
            next_item = (
                session.query(self.table_model)
                .order_by(id_field.asc())
                .filter(id_field > feature_id)
                .first()
            )
            feature['prev'] = (
                getattr(prev_item, self.id_field)
                if prev_item is not None
                else feature_id
            )
            feature['next'] = (
                getattr(next_item, self.id_field)
                if next_item is not None
                else feature_id
            )

        return feature
